"""Normal-reciprocal functionality
"""
import logging

import pandas as pd
import numpy as np

logger = logging.getLogger('__name__')


def _first(x):
    """return the first item of the supplied Series"""
    return x.iloc[0]


def average_repetitions(df, keys_mean):
    """average duplicate measurements. This requires that IDs and norrec labels
    were assigned using the *assign_norrec_to_df* function.

    Parameters
    ----------
    df
        DataFrame
    keys_mean: list
        list of keys to average. For all other keys the first entry will be
        used.
    """
    if 'norrec' not in df.columns:
        raise Exception(
            'The "norrec" column is required for this function to work!'
        )

    # Get column order to restore later
    cols = list(df.columns.values)

    keys_keep = list(set(df.columns.tolist()) - set(keys_mean))
    agg_dict = {x: _first for x in keys_keep}
    agg_dict.update({x: np.mean for x in keys_mean})
    for key in ('id', 'timestep', 'frequency', 'norrec'):
        if key in agg_dict:
            del(agg_dict[key])

    # average over duplicate measurements
    extra_dimensions_raw = ['id', 'norrec', 'frequency', 'timestep']
    extra_dimensions = [x for x in extra_dimensions_raw if x in df.columns]
    df = df.groupby(extra_dimensions).agg(agg_dict)
    df.reset_index(inplace=True)
    return df[cols]


def compute_norrec_differences(df, keys_diff):
    """DO NOT USE ANY MORE - DEPRECIATED!

    """
    raise Exception('This function is depreciated!')
    logger.info('computing normal-reciprocal differences')
    # df.sort_index(level='norrec')

    def norrec_diff(x):
        """compute norrec_diff"""
        if x.shape[0] != 2:
            return np.nan
        else:
            return np.abs(x.iloc[1] - x.iloc[0])

    keys_keep = list(set(df.columns.tolist()) - set(keys_diff))
    agg_dict = {x: _first for x in keys_keep}
    agg_dict.update({x: norrec_diff for x in keys_diff})
    for key in ('id', 'timestep', 'frequency'):
        if key in agg_dict:
            del(agg_dict[key])

    # for frequencies, we could (I think) somehow prevent grouping by
    # frequencies...
    df = df.groupby(('timestep', 'frequency', 'id')).agg(agg_dict)
    # df.rename(columns={'r': 'Rdiff'}, inplace=True)
    df.reset_index()
    return df


def _normalize_abmn(abmn):
    """return a normalized version of abmn
    """
    abmn_2d = np.atleast_2d(abmn)
    abmn_normalized = np.hstack((
        np.sort(abmn_2d[:, 0:2], axis=1),
        np.sort(abmn_2d[:, 2:4], axis=1),
    ))
    return abmn_normalized


def assign_norrec_to_df(df):
    """Determine normal-reciprocal pairs for a given dataframe.

    Parameters
    ----------
    df: pandas.DataFrame
        The data

    Returns
    -------
    df_new: pandas.DataFrame
        The data with two new columns: "id" and "norrec"

    """
    if df.shape[0] == 0:
        # empty dataframe, just return a copy
        return df.copy()

    c = df[['a', 'b', 'm', 'n']].values.copy()
    # unique injections
    cu = np.unique(c, axis=0)

    # now assign unique IDs to each config in normal and reciprocal
    running_index = 0
    normal_ids = {}
    reciprocal_ids = {}
    # loop through all configurations
    for i in range(0, cu.shape[0]):
        # normalize configuration
        cu_norm = _normalize_abmn(cu[i, :]).squeeze()
        if tuple(cu_norm) in normal_ids:
            continue

        # find pairs
        indices = np.where((
            # current electrodes
            (
                (
                    (cu[:, 0] == cu[i, 2]) & (cu[:, 1] == cu[i, 3])
                ) |
                (
                    (cu[:, 0] == cu[i, 3]) & (cu[:, 1] == cu[i, 2])
                )
            ) &
            # voltage electrodes
            (
                (
                    (cu[:, 2] == cu[i, 0]) & (cu[:, 3] == cu[i, 1])
                ) |
                (
                    (cu[:, 2] == cu[i, 1]) & (cu[:, 3] == cu[i, 0])
                )
            )
        ))[0]

        # we found no pair
        if len(indices) == 0:
            if not tuple(cu_norm) in normal_ids:
                if np.min(cu_norm[0:2]) < np.min(cu_norm[2:3]):
                    # treat as normal
                    normal_ids[tuple(cu_norm)] = running_index
                else:
                    reciprocal_ids[tuple(cu_norm)] = running_index
                running_index += 1
            continue

        # if len(indices) > 1:
        #     print('found more than one reciprocals')

        # normalize the first reciprocal
        cu_rec_norm = _normalize_abmn(cu[indices[0], :]).squeeze()

        # decide on normal or reciprocal
        # print('ABREC', cu_norm[0:2], cu_rec_norm[0:2])
        if np.min(cu_norm[0:2]) < np.min(cu_rec_norm[0:2]):
            # print('is normal')
            # normal
            normal_ids[tuple(cu_norm)] = running_index
            reciprocal_ids[tuple(cu_rec_norm)] = running_index
        else:
            normal_ids[tuple(cu_rec_norm)] = running_index
            reciprocal_ids[tuple(cu_norm)] = running_index
        running_index += 1

    # print(df.shape)
    # print(df.columns)
    # print('normal_ids', normal_ids)
    # print('reciprocal_ids', reciprocal_ids)
    # now convert the indices into a dataframe so we can use pd.merge
    # note that this code was previously written in another way, so the
    # conversion is quite cumbersome
    # at one point we need to rewrite everything here...
    df_nor = {item: key for key, item in normal_ids.items()}
    df_nor = pd.DataFrame(df_nor).T.reset_index().rename(
        {'index': 'id'}, axis=1)
    df_nor['norrec'] = 'nor'

    if len(normal_ids) > 0:
        df_nor.columns = ('id', 'a', 'b', 'm', 'n', 'norrec')
        df_nor2 = df_nor.copy()
        df_nor2.columns = ('id', 'b', 'a', 'm', 'n', 'norrec')
        df_nor3 = df_nor.copy()
        df_nor3.columns = ('id', 'b', 'a', 'n', 'm', 'norrec')
        df_nor4 = df_nor.copy()
        df_nor4.columns = ('id', 'a', 'b', 'n', 'm', 'norrec')
        df_ids = pd.concat(
            (
                df_nor,
                df_nor2,
                df_nor3,
                df_nor4,
            ),
            sort=True
        )
    else:
        df_ids = pd.DataFrame()

    if len(reciprocal_ids) > 0:
        df_rec = {item: key for key, item in reciprocal_ids.items()}
        df_rec = pd.DataFrame(df_rec).T.reset_index().rename(
            {'index': 'id'}, axis=1)
        df_rec['norrec'] = 'rec'
        df_rec.columns = ('id', 'a', 'b', 'm', 'n', 'norrec')
        df_rec2 = df_rec.copy()
        df_rec2.columns = ('id', 'b', 'a', 'm', 'n', 'norrec')
        df_rec3 = df_rec.copy()
        df_rec3.columns = ('id', 'b', 'a', 'n', 'm', 'norrec')
        df_rec4 = df_rec.copy()
        df_rec4.columns = ('id', 'a', 'b', 'n', 'm', 'norrec')

        df_ids = pd.concat(
            (
                df_ids,
                df_rec,
                df_rec2,
                df_rec3,
                df_rec4,
            ),
            sort=True
        )

    df_new = pd.merge(df, df_ids, how='left', on=('a', 'b', 'm', 'n'))
    df_new.rename(
        {'id_y': 'id',
         'norrec_y': 'norrec'
         }, axis=1,
        inplace=True
    )
    return df_new

    df_new[['a', 'b', 'm', 'n', 'id_y', 'norrec_y']]
    # x.iloc[[0, 1978], :]

    # now assign to all measurements
    for key, item in normal_ids.items():
        df.loc[
            ((df.a == key[0]) & (df.b == key[1]) &
             (df.m == key[2]) & (df.n == key[3])) |
            ((df.a == key[1]) & (df.b == key[0]) &
             (df.m == key[2]) & (df.n == key[3])) |
            ((df.a == key[0]) & (df.b == key[1]) &
             (df.m == key[3]) & (df.n == key[2])) |
            ((df.a == key[1]) & (df.b == key[0]) &
             (df.m == key[3]) & (df.n == key[2])),
            ('id', 'norrec')
        ] = (item, 'nor')
    for key, item in reciprocal_ids.items():
        df.loc[
            ((df.a == key[0]) & (df.b == key[1]) &
             (df.m == key[2]) & (df.n == key[3])) |
            ((df.a == key[1]) & (df.b == key[0]) &
             (df.m == key[2]) & (df.n == key[3])) |
            ((df.a == key[0]) & (df.b == key[1]) &
             (df.m == key[3]) & (df.n == key[2])) |
            ((df.a == key[1]) & (df.b == key[0]) &
             (df.m == key[3]) & (df.n == key[2])),
            ('id', 'norrec')
        ] = [item, 'rec']

    # cast norrec-column to string
    df['norrec'] = df['norrec'].astype(str)

    return df


def get_test_df():
    """Return a test dataframe suitable to test the normal-reciprocal functions
    """
    df = pd.DataFrame(
        [
            (1, 2, 3, 4, 10),
            (2, 1, 3, 4, 9),
            (1, 2, 4, 3, 8),
            (2, 1, 4, 3, 11),
            (3, 4, 1, 2, 12),
            (2, 3, 4, 5, 20),
            (4, 3, 3, 2, 17),

        ],
        columns=[
            'a',
            'b',
            'm',
            'n',
            'r',
        ]
    )
    return df


def get_test_df_advanced():
    """Return a test dataframe suitable to test the normal-reciprocal functions
    """
    df = pd.DataFrame(
        [
            (0, 0.1, 1, 2, 3, 4, 10),
            (0, 0.3, 3, 4, 1, 2, 12),
            (0, 0.4, 2, 3, 4, 5, 20),
            (0, 0.5, 4, 3, 3, 2, 17),
            (1, 0.1, 1, 2, 3, 4, 20),
            (1, 0.3, 1, 2, 3, 4, 20),
            (1, 0.3, 3, 4, 1, 2, 25),
            (1, 0.4, 2, 3, 4, 5, 30),
            (1, 0.5, 4, 3, 3, 2, 47),

        ],
        columns=[
            'timestep',
            'frequency',
            'a',
            'b',
            'm',
            'n',
            'r',
        ]
    )
    return df


def assign_norrec_diffs(df, diff_list):
    """Compute and write the difference between normal and reciprocal values
    for all columns specified in the diff_list parameter.

    Note that the DataFrame is directly written to. That is, it is changed
    during the call of this function. No need to use the returned object.

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe containing the data
    diff_list: list
        list of columns to compute differences for.

    Returns
    -------
    df_new: pandas.DataFrame
        The data with added columns
    """
    extra_dims = [
        x for x in ('timestep', 'frequency', 'id') if x in df.columns
    ]
    g = df.groupby(extra_dims)

    def subrow(row):
        if row.size == 2:
            return row.iloc[1] - row.iloc[0]
        else:
            return np.nan

    for diffcol in diff_list:
        # do nothing if the column does not exist
        if diffcol not in df.columns:
            continue
        diff = g[diffcol].agg(subrow).reset_index()
        # rename the column
        cols = list(diff.columns)
        cols[-1] = diffcol + 'diff'
        diff.columns = cols

        df = df.drop(
            cols[-1], axis=1, errors='ignore'
        ).merge(diff, on=extra_dims, how='outer')

    df = df.sort_values(extra_dims)
    return df


def test_norrec_assignments1():
    import reda.utils.norrec as redanr
    df = redanr.get_test_df()
    redanr.assign_norrec_to_df(df)
    df1 = redanr.average_repetitions(df, ['r', ])
    g = df1.groupby('id')
    diffs_R = g['r'].diff()

    def apply_nr_diff(row):
        return diffs_R.iloc[row['id']]
    df1['norrec_diff'] = df1.apply(apply_nr_diff, axis=1)


def test2():
    df = get_test_df_advanced()
    assign_norrec_to_df(df)
    df1 = average_repetitions(df, ['r', ])
    g = df1.groupby(['timestep', 'frequency', 'id'])

    def subrow(row):
        if row.size == 2:
            return row.iloc[1] - row.iloc[0]
        else:
            return np.nan

    diff = g['r'].agg(subrow).reset_index()
    cols = list(diff.columns)
    cols[-1] = 'Rdiff'
    diff.columns = cols
    df1 = df1.merge(diff)
    df1 = df1.sort_values(['timestep', 'frequency'])
    assert(
        df1.query(
            'timestep == 1 and a == 1 and b == 2 and m == 3 and n == 4 ' +
            'and frequency == 0.3'
        )['Rdiff'].values == 5.0
    )



""" Importer to load the unified data format used in pyGIMLi, BERT, and
dc2dinvres."""
import logging

import pandas as pd
import numpy as np

import reda

logger = logging.getLogger(__name__)


def import_ohm(filename, verbose=False, reciprocals=False, **kwargs):
    """Construct pandas data frame from BERT`s unified data format (.ohm).

    Parameters
    ----------
    filename : string
        File path to .ohm file
    verbose : bool, optional
        Enables extended debug output
    reciprocals : int, optional
        if provided, then assume that this is a reciprocal measurement where
        only the electrode cables were switched. The provided number N is
        treated as the maximum electrode number, and denotations are renamed
        according to the equation :math:`X_n = N - (X_a - 1)`

    Additional Parameters
    ---------------------
    shift_by_xyz : tuple|list|numpy.ndarray of size 1 or 2 or 3, optional
        If set, shift electrode positions by adding this vector Length of 1
        assumes that only the x coordinate of the vector differs from zero.
        Length of 2 assume a shift in (x,z) direction.

        This parameter is evaluated after the 'elecs_transform_reg_spacing_x'
        parameter - as such you can and must use real spacings in this case.

    Returns
    -------
    data : :class:`pandas.DataFrame`
       The measurement data
    elecs : :class:`pandas.DataFrame`
        Electrode positions (columns: X, Y, Z)
    topography : None
        No topography information is provided at the moment

    """
    if verbose:
        print(("Reading in %s... \n" % filename))
    file = open(filename)

    eleccount = int(file.readline().split("#")[0])
    elecs_str = file.readline().split("#")[1]
    elecs_dim = len(elecs_str.split())
    # elecs_ix = elecs_str.split()

    elecs = np.zeros((eleccount, elecs_dim), 'float')
    for i in range(eleccount):
        line = file.readline().split("#")[0]  # Account for comments
        elecs[i] = line.rsplit()

    datacount = int(file.readline().split("#")[0])
    data_str = file.readline().split("#")[1]
    data_dim = len(data_str.split())
    data_ix = data_str.split()

    _string_ = """
    Number of electrodes: %s
    Dimension: %s
    Coordinates: %s
    Number of data points: %s
    Data header: %s
    """ % (eleccount, elecs_dim, elecs_str, datacount, data_str)

    data = np.zeros((datacount, data_dim), 'float')
    for i in range(datacount):
        line = file.readline()
        data[i] = line.rsplit()

    file.close()

    data = pd.DataFrame(data, columns=data_ix)
    # rename columns to the reda standard
    data_reda = data.rename(
        index=str,
        columns={
            'rhoa': 'rho_a',
            # 'k': 'k',
            # 'u': 'U',
            # 'i': 'I'
        }
    )
    data_reda[['a', 'b', 'm', 'n']] = data_reda[['a', 'b', 'm', 'n']].astype(
        int
    )
    if ('r' not in data_reda.keys()) and \
       ('rho_a' in data_reda.keys() and 'k' in data_reda.keys()):
        data_reda['r'] = data_reda['rho_a'] / data_reda['k']
        logger.info(
            "Calculating resistance from apparent resistivity and "
            "geometric factors. (r = rhoa_ / k)")

    # rename electrode denotations
    if type(reciprocals) == int:
        logger.info('renumbering electrode numbers')
        data_reda[['a', 'b', 'm', 'n']] = reciprocals + 1 - data_reda[
            ['a', 'b', 'm', 'n']]

    if verbose:
        print((_string_))

    elec_mgr = reda.electrode_manager()
    elec_mgr.set_ordering_to_as_is_plus_one()

    elec_mgr.add_by_position(elecs)

    shift_by_xyz = kwargs.get('shift_by_xyz', None)
    if shift_by_xyz is not None:
        elec_mgr.shift_positions_xyz(shift_by_xyz)

    return data_reda, elec_mgr, None



"""Utility class for context managers and decorators
"""


def append_doc_of(fun):
    def decorator(f):
        f.__doc__ += fun.__doc__
        return f

    return decorator


def prepend_doc_of(fun):
    def decorator(f):
        f.__doc__ = fun.__doc__ + f.__doc__
        return f

    return decorator


class LogDataChanges():
    """Context manager that observes the DataFrame of a data container for
    changes in the number of rows.

    Examples
    --------

    >>> from reda.testing.containers import ERTContainer
    >>> from reda.containers.ERT import LogDataChanges
    >>> with LogDataChanges(ERTContainer):
    ...     # now change the data
    ...     ERTContainer.data.loc[0, "r"] = 22
    ...     ERTContainer.data.query("r < 10", inplace=True)
    >>> # ERTContainer.print_log()
    2... - root - INFO - Data size changed from 22 to 21

    """

    def __init__(
            self,
            container,
            filter_action='default',
            filter_query="",
    ):
        self.container = container
        self.logger = container.logger
        self.filter_action = filter_action
        self.data_size_before = None
        self.data_size_after = None
        self.filter_query = filter_query

    def __enter__(self):
        if self.container.data is None:
            self.data_size_before = 0
        else:
            self.data_size_before = self.container.data.shape[0]
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        # make sure we do not execute more code if an exception was raised in
        # the context
        # See:
        # https://docs.python.org/3/reference/datamodel.html#object.__exit__
        if(
            exc_type is not None or
            exc_value is not None or
                traceback is not None):
            return
        self.data_size_after = self.container.data.shape[0]
        self.logger.info(
            'Data sized changed from {0} to {1}'.format(
                self.data_size_before,
                self.data_size_after,
            ),
            extra={
                'filter_action': self.filter_action,
                'df_size_before': self.data_size_before,
                'df_size_after': self.data_size_after,
                'filter_query': self.filter_query,
            },
        )



"""This file contains transform functions for electrode numbering. If want to
apply custom electrode renumbering, use these functions as starting points to
design your own.
"""


class transform_electrodes_roll_along(object):
    """This function shifts all electrode numbers by a fixed offset, as
    commonly encountered for roll-a-long measurement schemes.

    This transformation should only be used for logical abmn transformations,
    i.e., in those cases were electrode positions are not available. Otherwise
    roll-along should be realized using coordinate transformations.
    """
    def __init__(self, shiftby=0):
        """
        Parameters
        ----------
        shiftby : int
            Shift electrode numbers (abmn) by this offset.
        """
        self.shiftby = shiftby

    def transform(self, data, electrodes, topography):
        if electrodes is not None:
            raise Exception(
                'Do not use the transform_electrodes_roll_along transform ' +
                'object if you have electrode positions!. Perform a ' +
                'roll-along alignment using the shiftby_xyz parameter of ' +
                'the importer functions.'
            )
        data[['a', 'b', 'm', 'n']] += self.shiftby
        return data, electrodes, topography



import numpy as np
"""
Convert between different representations for complex resistivity spectra

Basically we always have two parameters for each frequency. These two
parameters can be representated in various forms: conductivity/resistivity;
magnitude-phase/real-imaginary part.

Note that in this context it doesn't matter if we deal with conductivities or
conductances (same for resitivity and resistance).
"""


def split_data(data, squeeze=False):
    """
    Split 1D or 2D into two parts, using the last axis

    Parameters
    ----------
    data:
    squeeze : squeeze results to remove unnecessary dimensions
    """
    vdata = np.atleast_2d(data)
    nr_freqs = int(vdata.shape[1] / 2)
    part1 = vdata[:, 0:nr_freqs]
    part2 = vdata[:, nr_freqs:]
    if(squeeze):
        part1 = part1.squeeze()
        part2 = part2.squeeze()
    return part1, part2


def to_complex(mag, pha):
    complex_nr = mag * np.exp(1j / 1000 * pha.astype(np.float32))
    return complex_nr


def generic_magpha_to_reim(mag, pha):
    """
    Generically convert magnitude and phase to real and imaginary part using
    the formula :math:`mag \cdot exp(1j / 1000 * pha)`

    Thus it is suitable for resistivities, multiply conductivity phases with -1
    """
    complex_nr = to_complex(mag, pha)
    real_part = np.real(complex_nr)
    imag_part = np.imag(complex_nr)
    return real_part, imag_part

####################
# # from converter ##
####################


def from_ccomplex(data):
    cre = np.real(data)
    cim = np.imag(data)
    return cre, cim


def from_rcomplex(data):
    # rre = np.real(data)
    Y = 1.0 / data
    cre = np.real(Y)
    cim = np.imag(Y)
    return cre, cim


def from_cre_cim(data):
    cre, cim = split_data(data)
    return cre, cim


def from_cre_cmim(data):
    cre, cmim = split_data(data)
    return cre, -cmim


def from_cmag_cpha(data):
    cmag, cpha = split_data(data)
    cre, cim = generic_magpha_to_reim(cmag, cpha)
    return cre, cim


def from_log10rmag_rpha(data):
    rlog10mag, rpha = split_data(data)
    Z = to_complex(10 ** rlog10mag, rpha)
    Y = 1 / Z
    real_part = np.real(Y)
    imag_part = np.imag(Y)
    return real_part, imag_part


def from_lnrmag_rpha(data):
    rlnmag, rpha = split_data(data)
    Z = to_complex(np.exp(rlnmag), rpha)
    Y = 1 / Z
    real_part = np.real(Y)
    imag_part = np.imag(Y)
    return real_part, imag_part


def from_rmag_rpha(data):
    rmag, rpha = split_data(data)
    Z = to_complex(rmag, rpha)
    Y = 1 / Z
    real_part = np.real(Y)
    imag_part = np.imag(Y)
    return real_part, imag_part


def from_rre_rmim(data):
    rre, rmim = split_data(data)
    Z = rre - 1j * rmim
    Y = 1 / Z
    real_part = np.real(Y)
    imag_part = np.imag(Y)
    return real_part, imag_part


def from_rre_rim(data):
    rre, rim = split_data(data)
    Z = rre + 1j * rim
    Y = 1 / Z
    real_part = np.real(Y)
    imag_part = np.imag(Y)
    return real_part, imag_part


##################
# # to converter ##
##################
# converts from conductiviy re/im to various formats

def to_cre_cim(cre, cim):
    data = np.hstack((cre, cim))
    return data


def to_cre_cmim(cre, cim):
    cmim = -np.array(cim)
    data = np.hstack((cre, cmim))
    return data


def to_cmag_cpha(cre, cim):
    Y = cre + 1j * cim
    cmag = np.abs(Y)
    cpha = np.arctan2(cim, cre) * 1000
    return np.hstack((cmag, cpha))


def to_rre_rim(cre, cim):
    Y = cre + 1j * cim
    Z = 1 / Y
    real_p = np.real(Z)
    imag_p = np.imag(Z)
    return np.hstack((real_p, imag_p))


def to_rre_rmim(cre, cim):
    Y = cre + 1j * cim
    Z = 1 / Y
    real_p = np.real(Z)
    mimag_p = -np.imag(Z)
    return np.hstack((real_p, mimag_p))


def to_rmag_rpha(cre, cim):
    Y = cre + 1j * cim
    Z = 1 / Y
    real_p = np.real(Z)
    imag_p = np.imag(Z)
    mag = np.abs(Z)
    pha = np.arctan2(imag_p, real_p) * 1000
    return np.hstack((mag, pha))


def to_log10rmag_rpha(cre, cim):
    rmag_rpha = to_rmag_rpha(cre, cim)
    mag_slice = slice(0, int(rmag_rpha.shape[1] / 2))
    log10rmag_rpha = rmag_rpha
    log10rmag_rpha[:, mag_slice] = np.log10(rmag_rpha[:, mag_slice])
    return log10rmag_rpha


def to_lnrmag_rpha(cre, cim):
    rmag_rpha = to_rmag_rpha(cre, cim)
    mag_slice = slice(0, int(rmag_rpha.shape[1] / 2))
    lnrmag_rpha = rmag_rpha
    lnrmag_rpha[:, mag_slice] = np.log(rmag_rpha[:, mag_slice])
    return lnrmag_rpha


def to_ccomplex(cre, cim):
    return cre + 1j * cim


def to_rcomplex(cre, cim):
    Y = cre + 1j * cim
    Z = 1.0 / Y
    return Z


# store the converter functions in dicts
from_converters = {
    'lnrmag_rpha': from_lnrmag_rpha,
    'log10rmag_rpha': from_log10rmag_rpha,
    'rmag_rpha': from_rmag_rpha,
    'rre_rim': from_rre_rim,
    'rre_rmim': from_rre_rmim,
    'cmag_cpha': from_cmag_cpha,
    'cre_cim': from_cre_cim,
    'cre_cmim': from_cre_cmim,
    'ccomplex': from_ccomplex,
    'rcomplex': from_rcomplex,
}

to_converters = {
    'lnrmag_rpha': to_lnrmag_rpha,
    'log10rmag_rpha': to_log10rmag_rpha,
    'rmag_rpha': to_rmag_rpha,
    'rre_rim': to_rre_rim,
    'rre_rmim': to_rre_rmim,
    'cmag_cpha': to_cmag_cpha,
    'cre_cim': to_cre_cim,
    'cre_cmim': to_cre_cmim,
    'ccomplex': to_ccomplex,
    'rcomplex': to_rcomplex,
}


def convert(input_format, output_format, data, one_spectrum=False):
    """
    Convert from the given format to the requested format

    Parameters
    ----------
    input_format : format of input data (parameter 'data')
    output_format : format of output data
    data : numpy array containing data in specified input format
    one_spectrum : True|False, the input data comprises one spectrum. This
                   allows for an additional format of the data array.

    Possible input/output formats:
    ------------------------------

        'lnrmag_rpha'
        'log10rmag_rpha'
        'rmag_rpha'
        'rre_rim'
        'rre_rmim'
        'cmag_cpha'
        'cre_cim'
        'cre_cmim'
        'ccomplex'
        'rcomplex'

    Array format
    ------------

    data is either 1D or 2D. A 1D array correspond to one spectrum, with double
    the size of the frequencies (which are not needed for the conversion).
    Thus, the first halt either comprises a magnitude data, and the second one
    phase data, or the parts comprise real and imaginary parts.

    For the 2D case there exist two possibilities:

    First, if one_spectrum is False, then the first axis denotes the spectrum
    number, and each spectrum is stored on the second axis as described for the
    1D case.

    Second, if one_spectrum is True, and the first axis has the size two, then
    the axis denotes either magnitude (index 0) and phase (index 1), or real
    (index 0) and imaginary (index 1) parts. The second axis has the same size
    as there are frequencies.

    Internally we always convert to real part and imaginary part of
    conductivity, and then convert back to the output format.

    Return values are of the same dimensions as input variables.
    """
    if input_format == output_format:
        return data

    if input_format not in from_converters:
        raise KeyError('Input format {0} not known!'.format(input_format))

    if output_format not in to_converters:
        raise KeyError('Output format {0} not known!'.format(output_format))

    # internally we always work with the second axis of double the frequency
    # size
    if len(data.shape) == 2 and data.shape[0] == 2 and one_spectrum:
        work_data = np.hstack((data[0, :], data[1, :]))
        one_spec_2d = True
    else:
        work_data = data
        one_spec_2d = False

    cre, cim = from_converters[input_format](work_data)
    converted_data = to_converters[output_format](cre, cim)

    if one_spec_2d:
        part1, part2 = split_data(converted_data, True)
        converted_data = np.vstack((part1, part2))

    # reshape to input size (this should only be necessary for 1D data)
    if len(data.shape) == 1:
        converted_data = np.squeeze(converted_data)
    return converted_data



"""."""
import os
import functools
import logging
import copy
import json

import pandas as pd
import matplotlib.pylab as plt
import numpy as np

import reda
from reda.main.logger import LoggingClass
import reda.exporters.bert as reda_bert_export
import reda.exporters.crtomo as reda_crtomo_export
import reda.plotters.histograms as HS
import reda.plotters.pseudoplots as PS
import reda.utils.fix_sign_with_K as redafixK
import reda.utils.geometric_factors as redaK
from reda.utils import has_multiple_timesteps
from reda.utils.norrec import assign_norrec_diffs
from reda.utils.norrec import assign_norrec_to_df
from reda.utils.geometric_factors import apply_K

from reda.utils.decorators_and_managers import LogDataChanges

logger = logging.getLogger(__name__)


class ImportersBase(object):

    """
    Base class for all importer classes

    """

    def _add_to_container(
            self, data_to_add, electrode_positions=None, topography=None,
            metadata=None):
        """Add a given dataset to the container

        Parameters
        ----------
        data_to_add : pandas.DataFrame
            Measurement data in the form of a DataFrame, must adhere to the
            container contraints (i.e., must have all required columns)
        electrode_positions : :py:class:`reda.electrode_manager`|None
            If set, this electrode manager will be merged with any existing
            electrode positions, resulting in a unified electrode position
            assignment.
        topography : None
            Topography of the subsurface environment
        metadata : dict|None
            Metadata that is added to the metadata dict of the container
        """

        if electrode_positions is not None:
            if self.electrode_positions is None:
                self.electrode_positions = electrode_positions()
            else:
                logger.debug('Merging electrode positions of old and new data')
                elec_mgr = reda.electrode_manager()

                if isinstance(
                        electrode_positions,
                        reda.utils.electrode_manager.electrode_manager):
                    electrode_positions = electrode_positions()

                positions_aligned, abmn_old, abmn_addition = \
                    elec_mgr.align_assignments(
                        self.electrode_positions,
                        electrode_positions,
                        self.data[['a', 'b', 'm', 'n']],
                        data_to_add[['a', 'b', 'm', 'n']],
                    )
                self.data[['a', 'b', 'm', 'n']] = abmn_old
                data_to_add[['a', 'b', 'm', 'n']] = abmn_addition
                self.electrode_positions = positions_aligned

        if topography is not None:
            if self.topography is not None:
                # check if the provided topography is equal to those points
                # already saved.
                is_same = (topography == self.topography).all()
                if not is_same:
                    raise Exception(
                        'Merging of topographies is not implemented yet')
            else:
                self.topography = topography

        if metadata is not None:
            self.metadata.update(metadata)

        self._add_to_data(data_to_add)

    def _add_to_data(self, data):
        """Add data to the container

        Parameters
        ----------
        data : pandas.DataFrame
            Measurement data in the form of a DataFrame, must adhere to the
            container constraints (i.e., must have all required columns)
        """
        if self.data is None:
            self.data = data
        else:
            # we have existing data
            if 'timestep' in self.data.columns and 'timestep' in data.columns:
                # we must check that the types of the new data and the old data
                # timesteps match (we allow arbitrary types of timestep keys)
                check = data['timestep'].dtype == self.data['timestep'].dtype
                error_msg = ''.join((
                    'types of timestep-keys do not match: new:'
                    '{} old: {}'.format(
                        data['timestep'].dtype,
                        self.data['timestep'].dtype
                    )
                ))
                assert check, error_msg

            self.data = pd.concat(
                (self.data, data), ignore_index=True, sort=True
            )

        # clean any previous norrec-assignments
        if 'norrec' and 'id' in self.data.columns:
            self.data.drop(['norrec', 'id'], axis=1, inplace=True)
        self.data = assign_norrec_to_df(self.data)
        # note that columns not in the DataFrames are ignored, thus no problem
        # to include rho_a and rpha
        self.data = assign_norrec_diffs(self.data, ['r', 'rho_a', 'rpha'])

        # Put a, b, m, n in the front and ensure integers
        for col in tuple("nmba"):
            cols = list(self.data)
            cols.insert(0, cols.pop(cols.index(col)))
            self.data = self.data.iloc[:, self.data.columns.get_indexer(cols)]
            self.data[col] = self.data[col].astype(int)

        if 'timestep' in self.data:
            # make sure the timestep column is in the fifth position
            col_order = ['a', 'b', 'm', 'n', 'timestep']
            self.data = self.data.reindex(columns=(
                col_order +
                list(
                    [key for key in self.data.columns if key not in col_order]
                )
            ))

    def _describe_data(self, df=None):
        """
        Print statistics on a DataFrame by calling its .describe() function

        Parameters
        ----------
        df : None|pandas.DataFrame, optional
            if not None, use this DataFrame. Otherwise use self.data

        """
        if df is None:
            df_to_use = self.data
        else:
            df_to_use = df
        cols = []
        for test_col in self.required_columns:
            if test_col in df_to_use.columns:
                cols.append(test_col)
        print(df_to_use[cols].describe())

    def add_dataframe(self, data, timestep=None, **kwargs):
        """Add data to the container using another DataFrame

        Parameters
        ----------
        data : pandas.DataFrame
            Measurement data in the form of a DataFrame, must adhere to the
            container constraints (i.e., must have all required columns) and
            electrode positions already registered must match.
        """
        if timestep is not None:
            data['timestep'] = timestep
        self._add_to_data(data)

    def merge_container(self, container):
        """Merge the data and electrode positions from another container into
        this one.
        """
        logger.debug('Merging containers')
        print(type(self))

        self._add_to_container(
            container.data,
            container.electrode_positions, container.topography)


class ExportersBase(object):
    """This class provides wrappers for most of the exporter functions and is
    meant to be inherited by the ERT data container.

    See Also
    --------
    Importers
    """

    @functools.wraps(reda_bert_export.export_bert)
    def export_bert(self, filename):
        reda_bert_export.export_bert(self.data, self.electrode_positions,
                                     filename)

    @functools.wraps(export_bert)
    def export_pygimli(self, *args, **kargs):
        """Same as .export_bert"""
        self.export_bert(*args, **kargs)

    @functools.wraps(reda_crtomo_export.save_block_to_crt)
    def export_crtomo(self, filename, norrec='all', store_errors=False):
        """Export to CRTomo-compatible file"""
        reda_crtomo_export.save_block_to_crt(
            filename, self.data, norrec, store_errors
        )


class BaseContainer(LoggingClass, ImportersBase, ExportersBase):
    """."""

    def __init__(
            self, data=None, electrode_positions=None, topography=None,
            metadata=None, **kwargs):
        """
        Parameters
        ----------
        data : :py:class:`pandas.DataFrame`
            If not None, then the provided DataFrame is assumed to contain
            valid data previously prepared elsewhere. Please refer to the
            documentation for required columns.
        electrode_positions : :py:class:`pandas.DataFrame`
            If set, this is expected to be a DataFrame which contains electrode
            positions with columns: "x", "y", "z".
        topography : :py:class:`pandas.DataFrame`
            If set, this is expected to a DataFrame which contains topography
            information with columns: "x", "y", "z".

        """
        self.setup_logger(__name__)
        self.data = None
        if data is not None:
            self.check_dataframe(data)
            self._add_to_data(data)

        self.electrode_positions = electrode_positions
        self.topography = topography
        if metadata is None:
            self.metadata = {}
        else:
            assert isinstance(metadata, dict), "metadata must be a dict"
            self.metadata = metadata

    def save_metadata(self, filename):
        with open(filename, 'w') as fid:
            json.dump(self.metadata, fid)

    def load_metadata(self, filename):
        with open(filename, 'r') as fid:
            metadata = json.load(fid)
            self.metadata.update(metadata)

    def check_dataframe(self, dataframe):
        """Check the given dataframe for the required type and columns
        """
        if dataframe is None:
            return None

        # is this a DataFrame
        if not isinstance(dataframe, pd.DataFrame):
            raise Exception(
                'The provided dataframe object is not a pandas.DataFrame')

        required_columns = tuple("abmn")
        for column in required_columns:
            if column not in dataframe:
                raise Exception(
                    'Required column not in dataframe: {0}'.format(column))
        return dataframe

    def sub_filter(self, subset, filter, inplace=True):
        """Apply a filter to subset of the data

        Examples
        --------

        ::

            .subquery(
                'timestep == 2',
                'R > 4',
            )

        """
        # build the full query
        full_query = ''.join(('not (', subset, ') or not (', filter, ')'))
        with LogDataChanges(self, filter_action='filter', filter_query=filter):
            result = self.data.query(full_query, inplace=inplace)
        return result

    def filter(self, query, inplace=True):
        """Use a query statement to filter data. Note that you specify the data
        to be removed!

        Parameters
        ----------
        query : string
            The query string to be evaluated. Is directly provided to
            pandas.DataFrame.query
        inplace : bool
            if True, change the container dataframe in place (defaults to
            True). Otherwise, return a new ERT container which contains the
            filtered data.

        Returns
        -------
        result : :py:class:`reda.ERT`
            ERT container with filtered data

        """
        with LogDataChanges(self, filter_action='filter', filter_query=query):
            result = self.data.query(
                'not ({0})'.format(query),
                inplace=inplace,
            )
        if inplace:
            return self
        else:
            # create a new object of this type (e.g., ERT, IP, TDIP, ...)
            return self.__class__(data=result)

    def compute_K_analytical(self, spacing, **kwargs):
        """Compute geometrical factors over the homogeneous half-space with a
        constant electrode spacing
        """
        K = redaK.compute_K_analytical(self.data, spacing=spacing, **kwargs)
        self.data = redaK.apply_K(self.data, K, **kwargs)
        redafixK.fix_sign_with_K(self.data, **kwargs)

    def pseudosection(self, column='r', filename=None, log10=False, **kwargs):
        """Plot a pseudosection of the given column. Note that this function
        only works with dipole-dipole data at the moment.

        Parameters
        ----------
        column : string, optional
            Column to plot into the pseudosection, default: r
        filename : string, optional
            if not None, save the resulting figure directory to disc
        log10 : bool, optional
            if True, then plot values in log10, default: False
        **kwargs : dict
            all additional parameters are directly provided to
            :py:func:`reda.plotters.pseudoplots.PS.plot_pseudosection_type2`

        Returns
        -------
        fig : :class:`matplotlib.Figure`
            matplotlib figure object
            ax : :class:`matplotlib.axes`
            matplotlib axes object
        cb : colorbar object
            matplotlib colorbar object
        """
        fig, ax, cb = PS.plot_pseudosection_type2(
            self.data, column=column, log10=log10, **kwargs
        )
        if filename is not None:
            fig.savefig(filename, dpi=300)
        return fig, ax, cb

    def pseudosection_type3(
            self, column='r', filename=None, log10=False, **kwargs):
        """Plot a pseudosection of the given column. Note that this function
        only works with dipole-dipole data at the moment.

        Parameters
        ----------
        column : string, optional
            Column to plot into the pseudosection, default: r
        filename : string, optional
            if not None, save the resulting figure directory to disc
        log10 : bool, optional
            if True, then plot values in log10, default: False
        **kwargs : dict
            all additional parameters are directly provided to
            :py:func:`reda.plotters.pseudoplots.PS.plot_pseudosection_type2`

        Returns
        -------
        fig : :class:`matplotlib.Figure`
            matplotlib figure object
            ax : :class:`matplotlib.axes`
            matplotlib axes object
        cb : colorbar object
            matplotlib colorbar object
        """
        fig, ax, cb = PS.plot_pseudosection_type3(
            self.data, column=column, log10=log10, **kwargs
        )
        if filename is not None:
            fig.savefig(filename, dpi=300)
        return fig, ax, cb

    def histogram(self, column='r', filename=None, log10=False, **kwargs):
        """Plot a histogram of one data column"""
        return_dict = HS.plot_histograms(self.data, column)
        if filename is not None:
            return_dict['all'].savefig(filename, dpi=300)
        return return_dict

    def has_multiple_timesteps(self):
        """Return True if container has multiple timesteps."""
        return has_multiple_timesteps(self.data)

    def delete_measurements(self, row_or_rows):
        """Delete one or more measurements by index of the DataFrame.

        Resets the DataFrame index.

        Parameters
        ----------
        row_or_rows : int or list of ints
            Row numbers (starting with zero) of the data DataFrame (ert.data)
            to delete

        Returns
        -------

        None
        """
        self.data.drop(self.data.index[row_or_rows], inplace=True)
        self.data = self.data.reset_index()

    def to_configs(self):
        """Return a config object that contains the measurement configurations
        (a,b,m,n) from the data

        Returns
        -------
        config_obj : reda.ConfigManager
        """
        config_obj = reda.configs.configManager.ConfigManager()
        config_obj.add_to_configs(self.data[['a', 'b', 'm', 'n']].values)
        return config_obj

    def plot_electrode_positions_2d(self, ax=None, use_y_axis=False):
        """Create a 2D scatter plot for the electrode positions. By default use
        the x and z coordinates.

        Parameters
        ----------
        ax : matplotlib.axes, optional
            Axes object to plot to. If None, create a new figure
        use_y_axis : bool, optional
            If true then use the y coordinates instead of z coordinates for
            plotting

        Returns
        -------
        fig : matplotlib.Figure|None
            The Figure object related to ax. None if no electrode positions
            were registered yet.
        ax : matplotlib.Axes|None
            The axes object plotted to. None if no electrode positions were
            registered yet.
        """
        if self.electrode_positions is None:
            return None, None

        if ax is None:
            fig, ax_plot = plt.subplots(figsize=(16 / 2.54, 4 / 2.54))
        else:
            ax_plot = ax
            fig = ax.get_figure()

        elecs = reda.electrode_manager()
        elecs.add_fixed_assignments(self.electrode_positions)
        elecs.plot_coordinates_x_z_to_ax(ax_plot, use_y_axis=use_y_axis)
        if ax is None:
            # only touch the layout if we created the figure
            fig.tight_layout()
        return fig, ax

    def replace_electrode_positions(self, coordinates):
        """Replace the imported electrode coordinates by new ones. This
        function assumes and expects that the number of new coordinates is the
        same as the old ones (i.e., a simple replacement).

        If the input is a pandas DataFrame, assume that the columns x, y, z are
        present.



        Parameters
        ----------
        coordinates : str|numpy.ndarray|pandas.DataFrame
        """
        assert isinstance(self.electrode_positions, pd.DataFrame), \
            'There are no electrode positions to replace'

        if isinstance(coordinates, pd.DataFrame):
            assert 'x' in coordinates.columns
            assert 'y' in coordinates.columns
            assert 'z' in coordinates.columns
            coords = coordinates
        else:
            if isinstance(coordinates, str):
                # assume this is a file
                if os.path.isfile(coordinates):
                    coords_raw = np.loadtxt(coordinates)
                    print('raw')
                    print(coords_raw)
                else:
                    raise Exception(
                        'filename {} not found'.format(coordinates))
            elif isinstance(coordinates, np.ndarray):
                assert len(coordinates.shape) == 2, \
                    'array must be 2D: N x (1/2/3)'
                coords_raw = coordinates

            if coords_raw.shape[1] == 1:
                cols = ['x', ]
            elif coords_raw.shape[1] == 2:
                cols = ['x', 'z']
            elif coords_raw.shape[1] == 3:
                cols = ['x', 'y', 'z']
            coords = pd.DataFrame(coords_raw, columns=cols)
            print(coords)
            for key in ['x', 'y', 'z']:
                if key not in coords.columns:
                    coords[key] = 0

        assert coords.shape[0] == self.electrode_positions.shape[0]

        # now finally replace the coordinates
        self.electrode_positions[['x', 'y', 'z']] = coords[
            ['x', 'y', 'z']
        ].values

    def apply_k(self, k):
        """Apply geometric factors to the data

        """
        apply_K(self.data, k)

    def create_copy(self):
        """Create a copy if the object.
        This is useful to investigate different filtering strategies.
        """
        print('WARNING: Implementation and testing still in progress!!!!')

        new_obj = self.__class__()
        new_obj.data = copy.deepcopy(self.data)
        new_obj.topography = copy.deepcopy(self.topography)
        new_obj.electrode_positions = copy.deepcopy(
            self.electrode_positions)

        # what about the log?
        print('WARNING: Journal and log is not copied!')

        return new_obj

    def plot_topography_2d(self, ax=None, use_y_axis=False):
        """
        Plot topography points for x/z coordinates. If present, also plot
        electrode positions.

        Parameters
        ----------
        ax : matplotlib.Axes|None, optional
            If provided, plot into this axes object
        use_y_axis : bool, optional
            If true then use the y coordinates instead of z coordinates for
            plotting

        Returns
        -------
        fig : matplotlib.Figure|None
            figure object or None if not topography data is present
        ax : matplotlib.Axes|None
            axes object or None if not topography data is present
        """
        if self.topography is None:
            return None, None

        if ax is None:
            fig, ax = plt.subplots(figsize=(16 / 2.54, 5 / 2.54))
            we_created_the_figure = True
        else:
            fig = ax.get_figure()
            we_created_the_figure = False

        x_topo = self.topography['x']
        if use_y_axis:
            y_topo = self.topography['y']
        else:
            y_topo = self.topography['z']

        ax.plot(
            x_topo,
            y_topo,
            '.-',
            color='k',
            label='topography',
        )

        if self.electrode_positions is not None:
            x_el = self.electrode_positions['x']
            if use_y_axis:
                y_el = self.electrode_positions['y']
            else:
                y_el = self.electrode_positions['z']
            ax.scatter(
                x_el,
                y_el,
                label='electrodes',
            )
        ax.set_xlabel('x [m]')
        if use_y_axis:
            ax.set_ylabel('y [m]')
        else:
            ax.set_ylabel('z [m]')
        ax.legend()
        if we_created_the_figure:
            # only touch the layout if we created the figure
            fig.tight_layout()

        return fig, ax



"""
Fix signs in resistance measurements using the K factors. The sign of negative
resistance measurements can be switched if the geometrical factor is negative.
"""
import numpy as np


def fix_sign_with_K(dataframe, **kwargs):
    """Swap electrode denotations so that geometrical (K) factors become
    positive. Also, swap signs of all parameters affected by this process.

    Affected parameters, at the moment, are:

        * k
        * r
        * Vmn
        * Zt
        * rho_a
        * rpha

    Parameters
    ----------
    dataframe : pandas.DateFrame
        dataframe holding the data

    Returns
    -------
    dataframe : pandas.DateFrame
        the fixed dataframe

    """
    # check for required columns
    if 'k' not in dataframe or 'r' not in dataframe:
        raise Exception('k and r columns required!')

    indices_negative = (dataframe['k'] < 0) & (dataframe['r'] < 0)
    if np.where(indices_negative)[0].size == 0:
        # nothing to do here
        return dataframe

    # import IPython
    # IPython.embed()
    dataframe.iloc[
        indices_negative.values, dataframe.columns.get_indexer(['k', 'r'])
    ] *= -1

    # switch potential electrodes
    indices_switched_ab = indices_negative & (dataframe['a'] > dataframe['b'])
    indices_switched_mn = indices_negative & (dataframe['a'] < dataframe['b'])

    dataframe.iloc[
        indices_switched_ab.values, dataframe.columns.get_indexer(['a', 'b'])
    ] = dataframe.iloc[
        indices_switched_ab.values, dataframe.columns.get_indexer(['b', 'a'])
    ].values

    dataframe.iloc[
        indices_switched_mn.values, dataframe.columns.get_indexer(['m', 'n'])
    ] = dataframe.iloc[
        indices_switched_mn.values, dataframe.columns.get_indexer(['n', 'm'])
    ].values

    # switch sign of voltages
    if 'Vmn' in dataframe:
        dataframe.iloc[
            indices_negative.values, dataframe.columns.get_indexer(['Vmn'])
        ] *= -1

    if 'Zt' in dataframe:
        dataframe.iloc[
            indices_negative.values, dataframe.columns.get_indexer(['Zt'])
        ] *= -1

    if 'rho_a' in dataframe:
        # recompute to make sure the signs are consistent
        dataframe['rho_a'] = dataframe['r'] * dataframe['k']

    if 'Mx' in dataframe:
        # for now we have to loop here because we store numpy arrays within
        # each cell
        for index in np.where(indices_negative)[0]:
            dataframe.at[index, 'Mx'] *= -1

    # recompute phase values
    if 'rpha' in dataframe:
        if 'Zt' in dataframe:
            # recompute phase values
            dataframe['rpha'] = np.arctan2(
                np.imag(dataframe['Zt']), np.real(dataframe['Zt'])
            ) * 1e3
        else:
            raise Exception(
                'Recomputation of phase without Zt not implemented yet. ' +
                'See source code for more information'
            )
            """
            when the complex number is located in the fourth sector instead of
            the first, this corresponds to a phase shift by pi. For all values
            where magnitude < 0 and phase < 3000 mrad reverse this shift by pi
            by multiplying the complex number by -1:
            new_value = - 1 * (Magnitude * exp(i phi))
            Test this function by setting one measurement to
            -85.02069 -183.25 in radic column 6 and 7, should get -58 mrad when
            converted
            """
    # Make sure a, b, m, n stay integers.
    for col in ('a', 'b', 'm', 'n'):
        dataframe[col] = dataframe[col].astype(int)

    return dataframe


def test_fix_sign_with_K():
    """a few simple test cases
    """
    import numpy as np
    import pandas as pd
    configs = np.array((
        (1, 2, 3, 4, -10, -20),
        (1, 2, 4, 3, 10, 20),
    ))
    df = pd.DataFrame(configs, columns=['a', 'b', 'm', 'n', 'r', 'k'])
    df['rho_a'] = df['k'] * df['r']
    print('old')
    print(df)
    df = fix_sign_with_K(df)
    print('fixed')
    print(df)



"""
select and execute functions to compute geometric factors according to the
rcParams variable.
"""
import os
import pandas as pd
import numpy as np
import reda


def apply_K(df, k, **kwargs):
    """Apply the geometric factors to the dataset and compute (apparent)
    resistivities/conductivities

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe containing the data to which geometric factors shall be
        applied
    k : numpy.ndarray
        The geometric factors
    **kwargs: {}
        No kwargs at the moment

    Returns
    -------
    df : pandas.DataFrame
        The data with geometric factors applied
    """
    if 'k' not in df.columns:
        df['k'] = k

    if 'rho_a' not in df.columns:
        df['rho_a'] = df['r'] * df['k']

    if 'sigma_a' not in df.columns:
        df['sigma_a'] = 1.0 / df['rho_a']

    if 'Zt' in df.columns:
        df['rho_a_complex'] = df['Zt'] * df['k']
    return df


def compute_K_numerical(dataframe, settings=None, keep_dir=None):
    """Use a finite-element modeling code to infer geometric factors for meshes
    with topography or irregular electrode spacings.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        the data frame that contains the data
    settings : dict
        The settings required to compute the geometric factors. See examples
        down below for more information in the required content.
    keep_dir : path
        if not None, copy modeling dir here

    Returns
    -------
    K : :class:`numpy.ndarray`
        K factors (are also directly written to the dataframe)

    Examples
    --------
    ::

        settings = {
            'rho': 100,
            'elem': 'elem.dat',
            'elec': 'elec.dat',
            'sink_node': '100',
            '2D': False,
        }


    """
    inversion_code = reda.rcParams.get('geom_factor.inversion_code', 'crtomo')
    if inversion_code == 'crtomo':
        import reda.utils.geom_fac_crtomo as geom_fac_crtomo
        if keep_dir is not None:
            keep_dir = os.path.abspath(keep_dir)
        K = geom_fac_crtomo.compute_K(
            dataframe, settings, keep_dir)
    else:
        raise Exception(
            'Inversion code {0} not implemented for K computation'.format(
                inversion_code
            ))
    return K


def compute_K_analytical(dataframe, spacing, **kwargs):
    """Given an electrode spacing, compute geometrical factors using the
    equation for the homogeneous half-space (Neumann-equation)

    If a dataframe is given, use the column (a, b, m, n). Otherwise, expect an
    Nx4 arrray.

    Parameters
    ----------
    dataframe : pandas.DataFrame or numpy.ndarray
        Configurations, either as DataFrame
    spacing : float or numpy.ndarray
        distance between electrodes. If array, then these are the x-coordinates
        of the electrodes

    Returns
    -------
    K : numpy.ndarray
        The geometric factors corresponding to the provided configurations
    """
    if isinstance(dataframe, pd.DataFrame):
        configs = dataframe[['a', 'b', 'm', 'n']].values
    else:
        configs = dataframe

    r_am = np.abs(configs[:, 0] - configs[:, 2]) * spacing
    r_an = np.abs(configs[:, 0] - configs[:, 3]) * spacing
    r_bm = np.abs(configs[:, 1] - configs[:, 2]) * spacing
    r_bn = np.abs(configs[:, 1] - configs[:, 3]) * spacing

    K = 2 * np.pi / (1 / r_am - 1 / r_an - 1 / r_bm + 1 / r_bn)

    if isinstance(dataframe, pd.DataFrame):
        dataframe['k'] = K

    return K



# -*- coding: utf-8 -*-
"""This file set ups matplotlib plot functions for the whole package.

Import all necessary Matplotlib modules and set default options
To use this module, import * from it:

Examples
--------

>>> import reda.utils.mpl
>>> plt, mpl = reda.utils.mpl.setup()

"""

from reda.utils import which
latex = which("latex")


def setup(use_latex=False, overwrite=False):
    """Set up matplotlib imports and settings.

    Parameters
    ----------
    use_latex: bool, optional
        Determine if Latex output should be used. Latex will only be enable if
        a 'latex' binary is found in the system.
    overwrite: bool, optional
        Overwrite some matplotlib config values.

    Returns
    -------
    plt: :mod:`pylab`
        pylab module imported as plt
    mpl: :mod:`matplotlib`
        matplotlib module imported as mpl
    """
    # just make sure we can access matplotlib as mpl
    import matplotlib as mpl

    # general settings
    if overwrite:
        mpl.rcParams["lines.linewidth"] = 2.0
        mpl.rcParams["lines.markeredgewidth"] = 3.0
        mpl.rcParams["lines.markersize"] = 3.0
        mpl.rcParams["font.size"] = 12
        mpl.rcParams['mathtext.default'] = 'regular'
    if latex and use_latex:
        mpl.rcParams['text.usetex'] = True

        mpl.rc(
            'text.latex',
            preamble=''.join((
                #         r'\usepackage{droidsans}
                r'\usepackage[T1]{fontenc} ',
                r'\usepackage{sfmath} \renewcommand{\rmfamily}{\sffamily}',
                r'\renewcommand\familydefault{\sfdefault} ',
                r'\usepackage{mathastext} '
            ))
        )
    else:
        mpl.rcParams['text.usetex'] = False

    import matplotlib.pyplot as plt
    return plt, mpl


def mpl_get_cb_bound_next_to_plot(ax):
    """ Return the coordinates for a colorbar axes next to the provided axes
    object. Take into account the changes of the axes due to aspect ratio
    settings.

    Parts of this code are taken from the transforms.py file from matplotlib

    Important: Use only AFTER fig.subplots_adjust(...)

    Examples
    --------

    >>> import matplotlib as mpl
    >>> import matplotlib.pyplot as plt
    >>> from reda.utils.mpl import mpl_get_cb_bound_next_to_plot
    >>> fig, ax = plt.subplots()
    >>> fig.subplots_adjust(right=0.8)
    >>> plt_obj = ax.plot([1, 2, 3], [1, 2, 3], '.-')
    >>> cb_pos = mpl_get_cb_bound_next_to_plot(ax)
    >>> ax1 = fig.add_axes(cb_pos, frame_on=True)
    >>> cmap = mpl.cm.jet_r
    >>> norm = mpl.colors.Normalize(vmin=float(23), vmax=float(33))
    >>> cb1 = mpl.colorbar.ColorbarBase(
    ...     ax1,
    ...     cmap=cmap,
    ...     norm=norm,
    ...     orientation='vertical'
    ... )
    >>> cb1.locator = mpl.ticker.FixedLocator([23, 28, 33])
    >>> cb1.update_ticks()
    >>> # cb1.ax.artists.remove(cb1.outline)
    """
    position = ax.get_position()

    figW, figH = ax.get_figure().get_size_inches()
    fig_aspect = figH / figW
    box_aspect = ax.get_data_ratio()
    pb = position.frozen()
    pb1 = pb.shrunk_to_aspect(box_aspect, pb, fig_aspect).bounds

    ax_size = ax.get_position().bounds

    xdiff = (ax_size[2] - pb1[2]) / 2
    ydiff = (ax_size[3] - pb1[3]) / 2

    # the colorbar is set to 0.01 width
    sizes = [ax_size[0] + xdiff + ax_size[2] + 0.01,
             ax_size[1] + ydiff,
             0.01,
             pb1[3]]

    return sizes


def mpl_get_cb_bound_below_plot(ax):
    """ Return the coordinates for a colorbar axes below the provided axes
    object. Take into account the changes of the axes due to aspect ratio
    settings.

    Important: Use only AFTER fig.subplots_adjust(...)

    """
    position = ax.get_position()

    figW, figH = ax.get_figure().get_size_inches()
    fig_aspect = figH / figW
    box_aspect = ax.get_data_ratio()
    pb = position.frozen()
    pb1 = pb.shrunk_to_aspect(box_aspect, pb, fig_aspect).bounds

    ax_size = ax.get_position().bounds

    # xdiff = (ax_size[2] - pb1[2]) / 2
    # ydiff = (ax_size[3] - pb1[3]) / 2

    # the colorbar is set to 0.01 width
    sizes = [ax_size[0], ax_size[1], pb1[2], 0.03]

    return sizes



"""Spectral Electrical Impedance Tomography (sEIT) container

This container holds multi-frequency (spectral) imaging data, that is multiple
SIP/EIS spectra for different four-point spreads, usually used for subsequent
tomographic analysis.
"""
import gc
import os
from numbers import Number

import numpy as np
import pandas as pd

from reda.containers.BaseContainer import ImportersBase
from reda.containers.BaseContainer import BaseContainer
import reda.importers.eit_fzj as eit_fzj
import reda.importers.radic_sip256c as reda_sip256c
import reda.importers.crtomo as reda_crtomo_exporter
import reda.importers.mpt_das1 as mpt_das1
import reda.utils.eit_fzj_utils as eit_fzj_utils
import reda.utils.geometric_factors as geometric_factors
from reda.utils.fix_sign_with_K import fix_sign_with_K
import reda.eis.plots as eis_plot

from reda.utils.decorators_and_managers import append_doc_of
from reda.utils.decorators_and_managers import LogDataChanges
from reda.utils import opt_import

import reda.exporters.crtomo as exporter_crtomo

import reda.plotters.pseudoplots as PS
import reda.plotters.histograms as HS

import reda.utils.mpl
plt, mpl = reda.utils.mpl.setup()


class sEITImporters(ImportersBase):
    """This class provides wrappers for most of the importer functions, and is
    meant to be inherited by the data containers
    """
    @append_doc_of(reda_crtomo_exporter.load_seit_data)
    def import_crtomo(self, directory, frequency_file='frequencies.dat',
                      data_prefix='volt_', **kwargs):
        """CRTomo importer"""
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])

        # we get no electrode positions (dummy1) and no topography data
        # (dummy2)
        data, dummy1, dumm2 = reda_crtomo_exporter.load_seit_data(
            directory, frequency_file, data_prefix, **kwargs)
        if timestep is not None:
            data['timestep'] = timestep
        self._add_to_container(data)

        print('Summary:')
        self._describe_data(data)

    def import_sip256c(self, filename, settings=None, reciprocal=None,
                       **kwargs):
        """Radic SIP256c data import"""
        timestep = kwargs.get('timestep', None)
        if 'timestep' in kwargs:
            del (kwargs['timestep'])
        if settings is None:
            settings = {}
        # we get not electrode positions (dummy1) and no topography data
        # (dummy2)
        data, dummy1, dummy2 = reda_sip256c.parse_radic_file(
            filename, settings, reciprocal=reciprocal, **kwargs)
        if timestep is not None:
            data['timestep'] = timestep
        self._add_to_container(data)

        self.logger.info('Summary:')
        self._describe_data(data)

    def import_eit_fzj(self, filename, configfile, correction_file=None,
                       timestep=None, **kwargs):
        """EIT data import for FZJ Medusa systems"""
        # we get no electrode positions (dummy1) and no topography data
        # (dummy2)
        df_emd, dummy1, dummy2 = eit_fzj.read_3p_data(
            filename,
            configfile,
            **kwargs
        )
        if correction_file is not None:
            eit_fzj_utils.apply_correction_factors(df_emd, correction_file)

        if timestep is not None:
            df_emd['timestep'] = timestep

        self._add_to_container(df_emd)

        print('Summary:')
        self._describe_data(df_emd)

    def import_mpt_das1(self, filename, timestep=None, **kwargs):
        """Import MPT DAS-1 SIP data

        Parameters
        ----------
        filename : str
            Data file
        timestep : object, optional
            Timestep of the measurement, default: None
        """
        # check file type
        assert mpt_das1.get_measurement_type(filename) == 'sip'
        data, electrodes, topography = mpt_das1.import_das1_sip(
            filename
        )
        if timestep is not None:
            data['timestep'] = timestep

        self._add_to_container(data)
        self.electrode_positions = electrodes
        self.topography = topography

        print('Summary:')
        self._describe_data(data)


class sEIT(BaseContainer, sEITImporters):

    def __init__(self, dataframe=None):
        self.setup_logger()
        self.required_columns = [
            'a',
            'b',
            'm',
            'n',
            'r',
            'frequency',
            'rpha',
            # 'Zt',
        ]
        if dataframe is not None:
            self.check_dataframe(dataframe)
        # normal data (or full data, if reciprocals are not sorted
        self.data = None
        if dataframe is not None:
            self._add_to_container(dataframe)

    def check_dataframe(self, dataframe):
        """Check the given dataframe for the required columns
        """
        for column in self.required_columns:
            if column not in dataframe:
                raise Exception('Required column not in dataframe: {0}'.format(
                    column
                ))

    @property
    def abmn(self):
        return self.data.groupby(['a', 'b', 'm', 'n'])

    def subquery(self, subset, filter, inplace=True):
        """

        Examples
        --------

        ::

            subquery(
                'timestep == 2',
                'R > 4',
            )

        """
        # build the full query
        full_query = ''.join((
            'not (',
            subset,
            ') or not (',
            filter,
            ')',
        ))
        result = self.data.query(full_query, inplace=inplace)
        return result

    def query(self, query, inplace=True):
        """State what you want to keep

        """
        # TODO: add to queue
        result = self.data.query(query, inplace=inplace)
        return result

    def filter(self, query, inplace=True):
        """Use a query statement to filter data. Note that you specify the data
        to be removed!

        Parameters
        ----------
        query : string
            The query string to be evaluated. Is directly provided to
            pandas.DataFrame.query
        inplace : bool, optional (default: True)
            if True, change the container dataframe in place (defaults to True)

        Returns
        -------
        result : :py:class:`pandas.DataFrame`
            DataFrame that contains the result of the filter application

        """
        with LogDataChanges(self, filter_action='filter', filter_query=query):
            result = self.data.query(
                'not ({0})'.format(query),
                inplace=inplace,
            )
        return result

    def remove_frequencies(self, fmin, fmax):
        """Remove frequencies outside the provided range from the dataset.

        fmin and max will also be excluded.

        Parameters
        ----------
        fmin : float
            Minimal frequency to be excluded
        fmax : float
            Maximal frequency to be excluded
        """
        self.data.query(
            'frequency > {0} and frequency < {1}'.format(fmin, fmax),
            inplace=True
        )
        g = self.data.groupby('frequency')
        print('Remaining frequencies:')
        print(sorted(g.groups.keys()))

    def keep_frequencies(self, frequencies):
        """Keep only the requested frequencies, remove everything else.

        Parameters
        ----------
        frequencies : tuple
            Tuple of frequencies to keep.
        """
        assert isinstance(frequencies, tuple), \
            'Parameter frequencies must be a tuple!'
        self.data.query(
            'frequency in {}'.format(
                frequencies
            ),
            inplace=True,
        )

    def compute_K_analytical(self, spacing):
        """Assuming an equal electrode spacing, compute the K-factor over a
        homogeneous half-space.

        For more complex grids, please refer to the module:
        reda.utils.geometric_factors

        Parameters
        ----------
        spacing : float
            Electrode spacing

        """
        assert isinstance(spacing, Number)
        K = geometric_factors.compute_K_analytical(self.data, spacing)
        self.data = geometric_factors.apply_K(self.data, K)
        fix_sign_with_K(self.data)

    @append_doc_of(fix_sign_with_K)
    def fix_sign_with_K(self):
        """ """
        fix_sign_with_K(self.data)

    def scatter_norrec(self, filename=None, individual=False):
        """Create a scatter plot for all diff pairs

        Parameters
        ----------

        filename : string, optional
            if given, save plot to file
        individual : bool, optional
            if set to True, return one figure for each row

        Returns
        -------
        fig : matplotlib.Figure or list of :py:class:`matplotlib.Figure.Figure`
            objects the figure object
        axes : list of matplotlib.axes
            the individual axes

        """
        # if not otherwise specified, use these column pairs:
        std_diff_labels = {
            'r': 'rdiff',
            'rpha': 'rphadiff',
        }

        diff_labels = std_diff_labels

        # check which columns are present in the data
        labels_to_use = {}
        for key, item in diff_labels.items():
            # only use if BOTH columns are present
            if key in self.data.columns and item in self.data.columns:
                labels_to_use[key] = item

        g_freq = self.data.groupby('frequency')
        frequencies = list(sorted(g_freq.groups.keys()))

        if individual:
            figures = {}
            axes_all = {}
        else:
            Nx = len(labels_to_use.keys())
            Ny = len(frequencies)
            fig, axes = plt.subplots(
                Ny, Nx,
                figsize=(Nx * 2.5, Ny * 2.5)
            )

        for row, (name, item) in enumerate(g_freq):
            if individual:
                fig, axes_row = plt.subplots(
                    1, 2, figsize=(16 / 2.54, 6 / 2.54))
            else:
                axes_row = axes[row, :]
            # loop over the various columns
            for col_nr, (key, diff_column) in enumerate(
                    sorted(labels_to_use.items())):
                indices = np.where(~np.isnan(item[diff_column]))[0]
                ax = axes_row[col_nr]
                ax.scatter(
                    item[key],
                    item[diff_column],
                )
                ax.set_xlabel(key)
                ax.set_ylabel(diff_column)
                ax.set_title('N: {}'.format(len(indices)))
            if individual:
                fig.tight_layout()
                figures[name] = fig
                axes_all[name] = axes_row

        if individual:
            return figures, axes_all
        else:
            fig.tight_layout()
            return fig, axes

    def filter_incomplete_spectra(self, flimit=1000, percAccept=85):
        """Remove all data points that belong to spectra that did not retain at
        least **percAccept** percent of the number of data points.

        Parameters
        ----------
        flimit : float, optional (default: 1000)
            The frequency limit up to which the criterium will be applied. The
            value is included into the range (<= flimit)
        percAccept : float, optional (default: 85)
            The percentage of retained data points below which a spectrum is
            removed

        ..warning::

            This function does not honor additional dimensions (e.g.,
            timesteps) yet!

        """
        assert percAccept > 0 and percAccept < 100, \
            'percAccept must lie within ]0, 100)'

        def _retain_only_complete_spectra(item, fmax, acceptN):
            """Function called using pd.filter, applied to all spectra in the
            data set. Return true if the number of data points <= **fmax** in
            item is equal, or larger, than **acceptN**.

            Parameters
            ----------
            item : :py:class:`pandas.DataFrame`
                dataframe containing one spectrum
            fmax : float
                maximum frequency up to which data points are counted
            acceptN : int
                the number of data points required to pass this test

            Returns
            -------
            true : bool
                if enough data points are present
            false : bool
                if not enough data points are present
            """
            frequencies = item['frequency'].loc[item['frequency'] < fmax]
            fN = frequencies.size
            if fN >= acceptN:
                return True
            return False

        group_abmn = self.data.groupby(['a', 'b', 'm', 'n'])
        frequencies = np.array(
            list(sorted(self.data.groupby('frequency').groups.keys()))
        )
        assert frequencies.size > 0
        assert flimit >= frequencies.min() and flimit <= frequencies.max()
        Nlimit = len(np.where(frequencies <= flimit)[0])
        Naccept = np.ceil(Nlimit * percAccept / 100.0)
        self.data = group_abmn.filter(
            _retain_only_complete_spectra, fmax=flimit, acceptN=Naccept
        ).copy()

    def get_spectrum(self, nr_id=None, abmn=None, withK=False,
                     plot_filename=None):
        """
        Return a spectrum and its reciprocal counter part, if present in the
        dataset. Optimally, refer to the spectrum by its normal-reciprocal id.

        If the timestep column is present, then return dictionaries for normal
        and reciprocal data, with one sip_response object associated with each
        timestep.

        If the parameter plot_filename is specified, then plots will be created
        using the SIP objects.
        If multiple timesteps are present, then the parameter plot_filename
        will be used as a template, and the timesteps will be appended for each
        plot.

        Parameters
        ----------
        withK : bool
            If True, and the column "k" exists, then return an apparent
            spectrum with geometric factors included

        Returns
        -------
        spectrum_nor : :py:class:`reda.eis.plots.sip_response` or dict or None
            Normal spectrum. None if no normal spectrum is available
        spectrum_rec : :py:class:`reda.eis.plots.sip_response` or dict or None
            Reciprocal spectrum. None if no reciprocal spectrum is available
        fig : :py:class:`matplotlib.Figure.Figure`, optional
            Figure object (only if plot_filename is set)
        """
        assert nr_id is None or isinstance(nr_id, int)
        assert nr_id is None or abmn is None

        assert not withK or (withK and 'k' in self.data.columns)

        # Here are some problems with |dict|{} at end of 369 and 371

        # determine nr_id for given abmn tuple
        if abmn is not None:
            subdata = self.data.query(
                'a == {} and b == {} and m == {} and n == {}'.format(*abmn)
            ).sort_values('frequency')
            if subdata.shape[0] == 0:
                return None, None

            # determine the norrec-id of this spectrum
            nr_id = subdata['id'].iloc[0]

        # get spectra
        subdata_nor = self.data.query(
            'id == {} and norrec=="nor"'.format(nr_id)
        ).sort_values('frequency')

        subdata_rec = self.data.query(
            'id == {} and norrec=="rec"'.format(nr_id)
        ).sort_values('frequency')

        # create spectrum objects
        spectrum_nor = {}
        spectrum_rec = {}

        if subdata_nor.shape[0] > 0:
            # create a spectrum for each timestep
            if 'timestep' in subdata_nor.columns:
                g_nor_ts = subdata_nor.groupby('timestep')
                with_timesteps = True
            else:
                # create a dummy group
                g_nor_ts = subdata_nor.groupby('id')
                with_timesteps = False

            spectrum_nor = {}
            for timestep, item in g_nor_ts:
                if withK:
                    k = item['k']
                else:
                    k = 1

                if 'r_error' in item.columns:
                    rmag_error = item['r_error']
                else:
                    rmag_error = None

                if 'rpha_error' in item.columns:
                    rpha_error = item['rpha_error']
                else:
                    rpha_error = None

                spectrum_nor[timestep] = eis_plot.sip_response(
                    frequencies=item['frequency'].values,
                    rmag=item['r'] * k,
                    rpha=item['rpha'],
                    rmag_err=rmag_error,
                    rpha_err=rpha_error,
                )

        if subdata_rec.shape[0] > 0:
            if 'timestep' in subdata_rec.columns:
                g_rec_ts = subdata_rec.groupby('timestep')
                with_timesteps = True
            else:
                g_rec_ts = subdata_rec.groupby('id')
                with_timesteps = False

            spectrum_rec = {}
            for timestep, item in g_rec_ts:
                if withK:
                    k = item['k']
                else:
                    k = 1
                spectrum_rec[timestep] = eis_plot.sip_response(
                    frequencies=item['frequency'].values,
                    rmag=item['r'] * k,
                    rpha=item['rpha'],
                )

        def _reduce_dicts(dictA, dictB):
            if len(dictA) <= 1 and len(dictB) <= 1:
                # reduce
                if len(dictA) > 0:
                    dictA_reduced = [*dictA.values()][0]
                else:
                    dictA_reduced = (None, )
                if len(dictB) > 0:
                    dictB_reduced = [*dictB.values()][0]
                else:
                    dictB_reduced = (None, )

                return dictA_reduced, dictB_reduced
            else:
                # do nothing
                return dictA, dictB

        if plot_filename is not None:
            ending = plot_filename[-4:]

            all_timesteps = {
                k for d in (spectrum_nor, spectrum_rec) for k in d.keys()}
            pairs = {
                k: [d.get(k, None) for d in (
                    spectrum_nor, spectrum_rec
                )] for k in all_timesteps
            }
            for timestep, pair in pairs.items():
                if with_timesteps:
                    ts_suffix = '_ts_{}'.format(timestep)
                else:
                    ts_suffix = ''
                filename = plot_filename[:-4] + ts_suffix + ending

                if pair[0] is None and pair[1] is not None:
                    obj = pair[1]
                    title = 'a: {} b: {} m: {}: n: {}'.format(
                        *subdata_rec[['a', 'b', 'm', 'n']].values[0, :]
                    )
                else:
                    obj = pair[0]
                    title = 'a: {} b: {} m: {}: n: {}'.format(
                        *subdata_nor[['a', 'b', 'm', 'n']].values[0, :]
                    )

                fig = obj.plot(
                    filename,
                    reciprocal=pair[1],
                    return_fig=True,
                    title=title,
                )
            return [*_reduce_dicts(spectrum_nor, spectrum_rec), fig]

        return _reduce_dicts(spectrum_nor, spectrum_rec)

    def plot_all_spectra(self, outdir):
        r"""This is a convenience function to plot ALL spectra currently
        stored in the container. It is useful to asses whether data filters
        do perform correctly.

        Note that the function just iterates over all ids and plots the
        corresponding spectra, thus it is slow.

        Spectra a named using the format: \%.2i_spectrum_id_\{\}.png.

        Parameters
        ----------
        outdir : string
            Output directory to store spectra in. Created if it does not
            exist.
        """
        os.makedirs(outdir, exist_ok=True)

        abmn_id = self.data[['a', 'b', 'm', 'n', 'id']].groupby(
            'id'
        ).first().sort_values(['a', 'b', 'm', 'n'])

        for nr, (spec_id, abmn) in enumerate(abmn_id.iterrows()):
            print(
                'Plotting spectrum with id {} ({} / {})'.format(
                    spec_id, nr, abmn_id.shape[0]
                )
            )
            plot_filename = ''.join((
                outdir + os.sep,
                '{:04}_spectrum_{:02}_{:02}_{:02}_{:02}_id_{:04}.png'.format(
                    nr, *abmn, spec_id
                )
            ))
            spec_nor, spec_rec, spec_fig = self.get_spectrum(
                nr_id=spec_id,
                plot_filename=plot_filename
            )
            plt.close(spec_fig)
            gc.collect()

    def plot_pseudosections(self, column, filename=None, return_fig=False):
        """Create a multi-plot with one pseudosection for each frequency.

        Parameters
        ----------
        column : string
            which column to plot
        filename : None|string
            output filename. If set to None, do not write to file. Default:
            None
        return_fig : bool
            if True, return the generated figure object, also if filename is
            set. Default: False

        Returns
        -------
        fig : None|matplotlib.Figure
            if return_fig is set to True or filename is None, return the
            generated Figure object
        """
        assert column in self.data.columns

        g = self.data.groupby('frequency')
        fig, axes = plt.subplots(
            4, 2,
            figsize=(15 / 2.54, 20 / 2.54),
            sharex=True, sharey=True
        )
        for ax, (key, item) in zip(axes.flat, g):
            fig, ax, cb = PS.plot_pseudosection_type2(
                item, ax=ax, column=column
            )
            ax.set_title('f: {:.3f} Hz'.format(key))
        fig.tight_layout()
        if filename is not None:
            fig.savefig(filename, dpi=300)

        if return_fig or filename is None:
            return fig
        else:
            plt.close(fig)

    def export_to_crtomo_multi_frequency(self, directory, norrec='norrec'):
        """Export the sEIT data into data files that can be read by CRTomo.

        Parameters
        ----------
        directory : string
            output directory. will be created if required
        norrec : string (nor|rec|norrec)
            Which data to export. Default: norrec

        """
        exporter_crtomo.write_files_to_directory(
            self.data, directory, norrec=norrec
        )

    def export_to_crtomo_one_frequency(
            self, volt_file, frequency, norrec='norrec'):
        """Export one frequency into a CRTomo volt.dat file

        Parameters
        ----------
        volt_file : string
            output file. Will be overwritten if it exists
        frequency : float
            frequency to export
        norrec : str (nor|rec|norrec)
            Which data to export. Default: norrec
        """
        assert isinstance(frequency, float)
        frequency_data = self.data.query('frequency == {}'.format(frequency))
        exporter_crtomo.save_block_to_crt(
            volt_file, frequency_data, norrec=norrec
        )

    def export_to_crtomo_seit_manager(self, grid, norrec='norrec'):
        """Return a ready-initialized seit-manager object from the CRTomo
        tools. This function only works if the crtomo_tools are installed.

        WARNING: Not timestep aware!

        Parameters
        ----------
        grid : crtomo.crt_grid
            A CRTomo grid instance
        norrec : str (nor|rec|norrec)
            Which data to export. Default: norrec (all)

        """
        import crtomo
        subdata = self.data.query('norrec == "{}"'.format(norrec))
        g = subdata.groupby('frequency')
        seit_data = {}
        for name, item in g:
            print(name, item.shape, item.size)
            if item.shape[0] > 0:
                seit_data[name] = item[
                    ['a', 'b', 'm', 'n', 'r', 'rpha']
                ].values
        seit = crtomo.eitMan(grid=grid, seit_data=seit_data)
        return seit

    def export_to_crtomo_td_manager(self, grid, frequency, norrec='norrec'):
        """Return a ready-initialized tdman object from the CRTomo tools. Use
        the given frequency data to initialize it.

        WARNING: Not timestep aware!

        Parameters
        ----------
        grid : crtomo.crt_grid
            A CRTomo grid instance
        frequency : float
            The frequency to export data for
        norrec : str (nor|rec|norrec)
            Which data to export. Default: norrec (all)
        """
        subdata = self.data.query('norrec == "{}"'.format(norrec))
        import crtomo
        data = subdata.query('frequency == {}'.format(frequency))[
            ['a', 'b', 'm', 'n', 'r', 'rpha']
        ]
        tdman = crtomo.tdMan(grid=grid, volt_data=data)
        return tdman

    def export_to_ascii(
            file_frequency='frequencies.dat', file_data='data.dat'):
        pass

    def plot_histograms(
            self, column='r', primary_dim=None, filename=None, **kwargs):
        """Plot a histograms for all frequencies of one data column

        Parameters
        ----------
        column : str, optional
            data column to plot. defaults to "r" for resistance
        primary_dim : None|str
            ???
        filename : None|str
            Prefix for filename. Do not add a file ending here, as additional
            string will be appended here.
        **kwargs : dict
            ???

        TODO: Check saving to file for more than one secondary dimension
        Parameters
        ----------
        """
        dict_dimension, figs = HS.plot_histograms_extra_dims(
            self.data, column, primary_dim, **kwargs)
        if filename is not None:
            for key, item in figs.items():
                item.savefig(
                    filename + '_{}.jpg'.format(key).replace('_', '-'), dpi=300
                )
        return dict_dimension, figs

    @property
    def nr_frequencies(self):
        """Return the number of frequencies in the data set"""
        if self.data is None:
            return 0
        group_f = self.data.groupby('frequency')
        return group_f.ngroups

    @property
    def Nf(self):
        """Shortcut for self.nr_frequencies"""
        return self.nr_frequencies()

    @property
    def frequencies(self):
        """Return the frequencies contained in the data set"""
        if self.data is None:
            return 0
        frequencies = sorted(self.data.groupby('frequency').groups.keys())
        return frequencies

    @property
    def nr_timesteps(self):
        """Return the number of timesteps registered with this container"""
        if self.data is None or 'timestep' not in self.data:
            return 0
        group_ts = self.data.groupby('timestep')
        return group_ts.ngroups

    def correct_for_cable_inductances(self, inductance_matrix):
        if isinstance(inductance_matrix, np.ndarray):
            L_matrix = inductance_matrix
        else:
            # assume this is a filename
            ending = inductance_matrix[-3:]
            if ending == '.mat':
                import scipy.io
                mat = scipy.io.loadmat(inductance_matrix, squeeze_me=True)
                L_matrix = mat['L12']
            else:
                # assume numpy-saved .dat file
                L_matrix = np.loadtxt(inductance_matrix)

        def get_mutual_inductance(item):
            # determine single elements
            m_1_1 = L_matrix[item['a'] - 1, item['m'] - 1]
            m_1_2 = L_matrix[item['a'] - 1, item['n'] - 1]
            m_2_1 = L_matrix[item['b'] - 1, item['m'] - 1]
            m_2_2 = L_matrix[item['b'] - 1, item['n'] - 1]
            # Zhao  et al 2015
            M = (m_1_1 - m_1_2) - (m_2_1 - m_2_2)
            inductance = item['frequency'] * 2 * np.pi * M
            return inductance

        self.data['inductance'] = self.data.apply(
            get_mutual_inductance, axis=1
        )

        def apply_inductance(item):
            # If r and rpha are present, but Zt not, compute it first
            if 'Zt' not in item:
                item['Zt'] = item['r'] * np.exp(1j * item['rpha'] / 1000)

            # we need to change Zt, and then compute r, rpha
            Zt = item['Zt'] - 1j * item['inductance']
            # if item['Zt'].real > 0 and Zt.real < 0:
            #     # assume pi shift
            #     Zt *= -1

            rpha = np.arctan2(np.imag(Zt), np.real(Zt)) * 1000
            r = np.abs(Zt)

            return pd.Series({
                'Zt': Zt,
                'r': r,
                'rpha': rpha
            })

        tmpdf = self.data.apply(apply_inductance, axis=1)
        tmpdf['rpha'] = np.real(tmpdf['rpha']).astype(float)
        tmpdf['r'] = np.real(tmpdf['r']).astype(float)

        self.data[['Zt', 'r', 'rpha']] = tmpdf[['Zt', 'r', 'rpha']]

    def get_configs(self):
        """Return all unique (a,b,m,n) configurations in this container. Does
        not differentiate between frequencies.

        Returns
        -------
        unique_configs : Nx4 numpy.ndarray
            The configurations, 1. indexed (starting with 1)
        """
        return np.array(list(self.abmn.groups.keys()))

    def keep_only_configs(self, configs_to_keep, frequency=None):
        """Keep only the supplied configs. If frequency is provided, only this
        frequency is filtered and all other frequencies are left as is.

        Parameters
        ----------

        """
        if frequency is not None:
            g = self.data.groupby(['frequency', 'a', 'b', 'm', 'n'])
        else:
            g = self.data.groupby(['a', 'b', 'm', 'n'])

        def filter_groups_abmn(x):
            if frequency is not None:
                if x['frequency'].iloc[0] != frequency:
                    return True

            abmn = x[['a', 'b', 'm', 'n']].values[0]

            keep_config = np.any(
                np.all(abmn == configs_to_keep, axis=1)
            )

            if keep_config:
                return True
            return False

        self.data = g.filter(filter_groups_abmn)

    def debye_decomposition_one_spectrum(
            self, abmn, spectrum, gen_plots_in_dir=None):
        """Conduct a Debye Decomposition on each spectrum. Save certain
        parameters in sEIT.data columns and/or return fit data
        """
        opt_import('lib_dd')
        import lib_dd.decomposition.ccd_single as ccd_single
        import lib_dd.config.cfg_single as cfg_single

        print(abmn)

        # set options using this dict-like object
        config = cfg_single.cfg_single()
        config['frequency_file'] = spectrum['frequency'].values
        rmag_rpha = np.hstack(
            (
                spectrum['r'],
                spectrum['rpha'],
            )
        )
        # print('rmag_rpha')
        # print(rmag_rpha)
        config['data_file'] = rmag_rpha
        ccd_obj = ccd_single.ccd_single(config)
        ccd_obj.fit_data()
        last_it = ccd_obj.results[0].iterations[-1]

        if gen_plots_in_dir is not None:
            pwd = os.getcwd()
            os.makedirs(gen_plots_in_dir, exist_ok=True)
            os.chdir(gen_plots_in_dir)
            filename = 'plot_{}.jpg'.format(
                abmn.values[0, :]
            )
            print('FILENAME', filename)
            last_it.plot(filename=filename)
            os.chdir(pwd)

        return last_it

    def debye_decomposition_all_spectra(self, gen_plots_in_dir=None):
        def apply_dd(spectrum):
            # print(abmn)
            # return 1
            last_it = self.debye_decomposition_one_spectrum(
                spectrum[['a', 'b', 'm', 'n']],
                spectrum,
                gen_plots_in_dir,
            )
            rms_re, rms_im = last_it.rms_values['rms_re_im_noerr']
            return rms_im

        r = self.abmn.apply(apply_dd)
        # for abmn, spectrum in self.abmn:
        #     last_it = self.debye_decomposition_one_spectrum(abmn, spectrum)
        #     rms_re, rms_im = last_it.rms_values['rms_re_im_noerr']
        #     print(rms_re)
        #     print(rms_im)
        #     exit()
        import IPython
        IPython.embed()



"""
Test reda.importers.crtomo functionality

"""
import reda.importers.crtomo as crtomo_importer
import os
import tempfile
import numpy as np


def test_load_complex_data():
    with tempfile.TemporaryDirectory() as directory:
        filename = directory + os.sep + 'volt.dat'
        with open(filename, 'w') as fid:
            fid.write('2\n')
            fid.write('10002 40003 3 -10\n')
            fid.write('50006 90008 6 -11.5\n')
        data = crtomo_importer.load_mod_file(filename)
        print(data)
        assert np.all(data['r'] == [3, 6]), 'Resistances do not match input'

        assert np.all(
            data['rpha'] == [-10.0, -11.5]
        ), 'Phases do not match input'

        assert np.all(
            data['a'] == [1, 5]
        ), 'A electrodes do not match'
        assert np.all(
            data['b'] == [2, 6]
        ), 'B electrodes do not match'
        assert np.all(
            data['m'] == [4, 9]
        ), 'M electrodes do not match'
        assert np.all(
            data['n'] == [3, 8]
        ), 'N electrodes do not match'


def test_load_resistance_data():
    with tempfile.TemporaryDirectory() as directory:
        filename = directory + os.sep + 'volt.dat'
        with open(filename, 'w') as fid:
            fid.write('2\n')
            fid.write('10002 40003 3\n')
            fid.write('50006 90008 6\n')
        data = crtomo_importer.load_mod_file(filename)
        print(data)
        assert np.all(data['r'] == [3, 6]), 'Resistances do not match input'

        assert np.all(
            data['a'] == [1, 5]
        ), 'A electrodes do not match'
        assert np.all(
            data['b'] == [2, 6]
        ), 'B electrodes do not match'
        assert np.all(
            data['m'] == [4, 9]
        ), 'M electrodes do not match'
        assert np.all(
            data['n'] == [3, 8]
        ), 'N electrodes do not match'

        assert 'rpha' not in data.columns, 'column rpha should not be present'
        assert 'Zt' not in data.columns, 'column Zt should not be present'
    return data


if __name__ == '__main__':
    data = test_load_resistance_data()



"""Importer for RES2DINV files

Please note that this importer is very rudimentary at the moment. We need test
data and real usage to improve upon this.

"""
from io import StringIO
import logging

import pandas as pd

from reda.utils.electrode_manager import electrode_manager

logger = logging.getLogger(__name__)


def import_res2dinv(filename, **kwargs):
    """ Read a RES2DINV-style

    Parameters
    ----------

    Returns
    -------
    data : :py:class:`pandas.DataFrame`
        Contains the measurement data
    elec_mgr : :py:class:`reda.utils.electrode_manager.electrode_manager`
        Electrode manager that manages the electrode positions
    topography : None
        No topography information is contained in the text files, so we always
        return None
    """
    # each type is read by a different function
    importers = {
        # general array type
        11: _read_general_type,
    }

    file_type, content = _read_file(filename)

    if file_type not in importers:
        raise Exception(
            'type of RES2DINV data file not recognized: {0}'.format(file_type)
        )

    header, data, elec_mgr = importers[file_type](content)

    topography = None
    return data, elec_mgr, topography


def _read_file(filename):
    """
    Read a res2dinv-file and return the header

    Parameters
    ----------
    filename : str
        Data filename

    Returns
    ------
    type : int
        type of array extracted from header
    file_data : :py:class:`StringIO.StringIO`
        content of file in a StringIO object
    """
    # read data from file
    with open(filename, 'r') as fid2:
        abem_data_orig = fid2.read()

    # for easy access, store the data in a StringIO container
    fid = StringIO()
    fid.write(abem_data_orig)
    fid.seek(0)

    # determine type of array
    fid.readline()
    fid.readline()

    file_type = int(fid.readline().strip())

    # reset file pointer
    fid.seek(0)
    return file_type, fid


def _read_general_type(content):
    """Read a type 11 (general type) RES2DINV data block

    Parameters
    ----------
    content : :py:class:`StringIO.StringIO`
        Content of data file

    Returns
    -------

    """
    header_raw = []
    index = 0
    while index < 9:
        line = content.readline()
        # filter comments
        if line.startswith('//'):
            continue
        else:
            header_raw.append(line.strip())
            index += 1
    # parse header
    header = {
        'name': header_raw[0],
        # unit is meters?
        'unit_spacing': float(header_raw[1]),
        'array_type': int(header_raw[2]),
        'type2': int(header_raw[3]),
        'type_of_measurements': int(header_raw[5]),
        'nr_measurements': int(header_raw[6]),
        'type_of_x_location': int(header_raw[7]),
        'ip_data': int(header_raw[8]),
    }

    if header['type_of_measurements'] == 0:
        raise Exception('Reading in app. resistivity not supported yet')

    df = pd.read_csv(
        content,
        delim_whitespace=True,
        header=None,
        names=(
            'nr_elecs',
            'x1',
            'z1',
            'x2',
            'z2',
            'x3',
            'z3',
            'x4',
            'z4',
            'value',
        ),
        nrows=header['nr_measurements'],
    )

    # type of x-location
    # 0: true horizontal distance
    # 1: true horizontal distance
    # 2: surface distance

    elec_mgr = electrode_manager()

    # for now ignore the z coordinates and compute simple electrode denotations
    df['a'] = df['x1'] / header['unit_spacing'] + 1
    df['b'] = df['x2'] / header['unit_spacing'] + 1
    df['m'] = df['x3'] / header['unit_spacing'] + 1
    df['n'] = df['x4'] / header['unit_spacing'] + 1

    # for now assume value in resistances
    df['r'] = df['value']

    # ABMN are integers
    df['a'] = df['a'].astype(int)
    df['b'] = df['b'].astype(int)
    df['m'] = df['m'].astype(int)
    df['n'] = df['n'].astype(int)

    # remove any nan values
    df.dropna(axis=0, subset=['a', 'b', 'm', 'n', 'r'], inplace=True)

    # drop unused columns
    df.drop(
        [
            'nr_elecs',
            'x1', 'z1',
            'x2', 'z2',
            'x3', 'z3',
            'x4', 'z4',
            'value',
        ], axis=1, inplace=True
    )
    return header, df, elec_mgr



"""
Various ways to compute pseudo-positions of given electrode configurations.
"""
import numpy as np


def get_xy_simple_dipole_dipole(dataframe, spacing=1, indices=None):
    """For each configuration indicated by the numerical index array, compute
    (x,z) pseudo locations based on the paper from XX.

    All positions are computed for indices=None.
    """
    if indices is None:
        indices = slice(None)
    abmn = dataframe.iloc[
        indices, dataframe.columns.get_indexer(['a', 'b', 'm', 'n'])
    ].values
    posx = np.mean(abmn[:, 0:4], axis=1)
    posz = np.abs(
        np.min(abmn[:, 0:2], axis=1) - np.max(abmn[:, 2:4], axis=1)
    ) * -0.192

    # scale the positions with the electrode spacing
    posx *= spacing
    posz *= spacing
    print(abmn.shape, posx.shape)
    print('posxz', np.vstack((abmn.T, posx, posz)).T)
    return posx, posz



# *-* coding: utf-8 *-*
import numpy as np

from reda.eis.units import get_label
from reda.eis.convert import convert

import reda.utils.mpl
plt, mpl = reda.utils.mpl.setup()


class sip_response():
    """Hold one EIS/SIP spectrum and return it in various formats
    """
    def __init__(self, frequencies, rcomplex=None, ccomplex=None,
                 rmag=None, rpha=None, rmag_err=None, rpha_err=None):
        """

        WARNING: rmag_err and rpha_err are only used for plotting. At this
        point no error propagation is done.

        Parameters
        ----------
        frequencies : :class:`numpy.ndarray`
            Array of size N containing N frequencies in ascending order
        rcomplex : :class:`numpy.ndarray`, optional
            Complex values resistance/resistivity values (size N)
        ccomplex : :class:`numpy.ndarray`, optional
            Complex values conductance/conductivity values (size N)
        rmag : :class:`numpy.ndarray`, optional
            Real valued resistance/resistivity magnitude values (size N)
        rpha : :class:`numpy.ndarray`, optional
            Real valued resistance/resistivity phase values (size N)
        rmag_err : :class:`numpy.ndarray`, optional
            Magnitude error
        rpha_err : :class:`numpy.ndarray`, optional
            Phase error

        """
        if rcomplex is None and ccomplex is None and (
                rmag is None or rpha is None):
            raise Exception('One initialization array is allowed!')
        if rcomplex is not None and ccomplex is not None:
            raise Exception('Only one initialization array is allowed!')

        self.frequencies = frequencies

        if rcomplex is not None:
            self.rcomplex = rcomplex
            self.ccomplex = convert('rcomplex', 'ccomplex', rcomplex)
        elif ccomplex is not None:
            self.ccomplex = ccomplex
            self.rcomplex = convert('ccomplex', 'rcomplex', ccomplex)
        elif rmag is not None and rpha is not None:
            self.rcomplex = rmag * np.exp(1j * rpha / 1000.0)
            self.ccomplex = convert('rcomplex', 'ccomplex', self.rcomplex)

        self.rmag_error = None
        self.rpha_error = None
        if rmag_err is not None:
            self.rmag_error = rmag_err
        if rpha_err is not None:
            self.rpha_error = rpha_err

        self.rmag = np.abs(self.rcomplex)
        self.rpha = np.arctan2(
            np.imag(self.rcomplex),
            np.real(self.rcomplex)
        ) * 1000
        self.cmag = np.abs(self.ccomplex)
        self.cpha = np.arctan2(
            np.imag(self.ccomplex),
            np.real(self.ccomplex)
        ) * 1000

        self.rmag_rpha = np.vstack((self.rmag, self.rpha)).T
        self.cmag_cpha = np.vstack((self.cmag, self.cpha)).T

        self.rre = np.real(self.rcomplex)
        self.rim = np.imag(self.rcomplex)
        self.cre = np.real(self.ccomplex)
        self.cim = np.imag(self.ccomplex)

        self.rre_rim = np.vstack((self.rre, self.rim)).T
        self.cre_cim = np.vstack((self.cre, self.cim)).T

    def to_one_line(self, array):
        """Flatten the array to one dimension using the 'F' (Fortran) style and
        return a 2D array
        """
        return np.atleast_2d(array.flatten(order='F'))

    def _add_labels(self, axes, dtype):
        """Given a 2x2 array of axes, add x and y labels

        Parameters
        ----------
        axes: numpy.ndarray, 2x2
            A numpy array containing the four principal axes of an SIP plot
        dtype: string
            Can be either 'rho' or 'r', indicating the type of data that is
            plotted: 'rho' stands for resistivities/conductivities, 'r' stands
            for impedances/condactances

        Returns
        -------
        None
        """
        for ax in axes[1, :].flat:
            ax.set_xlabel('frequency [Hz]')

        if dtype == 'rho':
            axes[0, 0].set_ylabel(r'$|\rho| [\Omega m]$')
            axes[0, 1].set_ylabel(r'$-\phi [mrad]$')
            axes[1, 0].set_ylabel(r"$\sigma' [S/m]$")
            axes[1, 1].set_ylabel(r"$\sigma'' [S/m]$")
        elif dtype == 'r':
            axes[0, 0].set_ylabel(r'$|R| [\Omega]$')
            axes[0, 1].set_ylabel(r'$-\phi [mrad]$')
            axes[1, 0].set_ylabel(r"$Y' [S]$")
            axes[1, 1].set_ylabel(r"$Y'' [S]$")
        else:
            raise Exception('dtype not known: {}'.format(dtype))

    def _plot(self, title=None, reciprocal=None, limits=None, dtype='rho',
              **kwargs):
        """Standard plot of spectrum

        Parameters
        ----------
        title : str|None, optional
            Title of plot
        reciprocal : sip_response object|None, optional
            If provided, plot this spectrum with another color
        limits : dict|None, optional
            used to set ylimits of the plots. Possible entries: rmag_min,
            rmag_max, rpha_min, rpha_max, cre_min, cre_max, cim_min, cim_max
        dtype : str, optional
            Possible values: [rho|R]. Determines the label types. 'rho':
                resistivity/conductivity, 'r': resistance/conductance
        label_nor : str
            label for normal data (default: "normal")
        label_rec : str
            label for reciprocal data (default: "reciprocal")

        Additional Parameters
        ---------------------


        Returns
        -------
        fig : figure object
            the generated matplotlib figure
        axes : list
            matplotlib axes objects
        """
        if limits is None:
            limits = {}

        fig, axes = plt.subplots(
            2, 2, figsize=(15 / 2.54, 6 / 2.54), sharex=True
        )
        if title is not None:
            fig.suptitle(title)

        # resistivity magnitude
        if limits is None:
            limits = {}

        ax = axes[0, 0]
        ax.semilogx(
            self.frequencies, self.rmag, '.-', color='k',
            label=kwargs.get('label_nor', 'normal'),
        )
        if self.rmag_error is not None:
            ax.fill_between(
                self.frequencies,
                self.rmag - self.rmag_error,
                self.rmag + self.rmag_error,
                alpha=0.5,
            )

        # resistivity phase
        ax = axes[0, 1]
        ax.semilogx(self.frequencies, -self.rpha, '.-', color='k')
        if self.rpha_error is not None:
            ax.fill_between(
                self.frequencies,
                -self.rpha - self.rpha_error,
                -self.rpha + self.rpha_error,
                alpha=0.5,
            )

        # note the switch of _min/_max because we change the sign while
        # plotting
        ymin = limits.get('rpha_max', None)
        if ymin is not None:
            ymin *= -1
        ymax = limits.get('rpha_min', None)
        if ymax is not None:
            ymax *= -1
        ax.set_ylim(
            ymin,
            ymax,
        )

        # conductivity real part
        ax = axes[1, 0]
        ax.loglog(self.frequencies, self.cre, '.-', color='k')
        ax.set_ylim(
            limits.get('cre_min', None),
            limits.get('cre_max', None)
        )

        # conductivity imaginary part
        ax = axes[1, 1]
        ax.loglog(self.frequencies, self.cim, '.-', color='k')
        ax.set_ylim(
            limits.get('cim_min', None),
            limits.get('cim_max', None)
        )

        self._add_labels(axes, dtype)

        for ax in axes.flatten()[0:2]:
            ax.xaxis.set_major_locator(mpl.ticker.LogLocator(numticks=5))
            ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(5))

        for ax in axes.flatten()[2:]:
            ax.xaxis.set_major_locator(mpl.ticker.LogLocator(numticks=5))
            ax.yaxis.set_major_locator(mpl.ticker.LogLocator(numticks=5))

        fig.tight_layout()
        # plot reciprocal spectrum
        if reciprocal is not None:
            axes[0, 0].semilogx(
                reciprocal.frequencies,
                reciprocal.rmag,
                '.-',
                color='k',
                linestyle='dotted',
                label=kwargs.get('label_rec', 'reciprocal'),
            )
            axes[0, 0].relim()
            axes[0, 0].autoscale()
            ax.set_ylim(
                limits.get('rmag_min', None),
                limits.get('rmag_max', None)
            )

            axes[0, 1].semilogx(
                reciprocal.frequencies,
                -reciprocal.rpha,
                '.-',
                color='k',
                linestyle='dotted',
            )
            axes[0, 1].relim()
            axes[0, 1].autoscale()
            ymin = limits.get('rpha_max', None)
            if ymin is not None:
                ymin *= -1
            ymax = limits.get('rpha_min', None)
            axes[0, 1].set_ylim(ymin, ymax)

            axes[1, 0].loglog(
                reciprocal.frequencies,
                reciprocal.cre,
                '.-',
                color='k',
                linestyle='dotted',
            )

            axes[1, 0].relim()
            axes[1, 0].autoscale()
            axes[1, 0].set_ylim(
                limits.get('cre_min', None),
                limits.get('cre_max', None)
            )
            axes[1, 1].loglog(
                reciprocal.frequencies,
                reciprocal.cim,
                '.-',
                color='k',
                linestyle='dotted',
            )
            axes[1, 1].relim()
            axes[1, 1].autoscale()
            axes[1, 1].set_ylim(
                limits.get('cim_min', None),
                limits.get('cim_max', None)
            )

            fig.subplots_adjust(
                bottom=0.3,
            )

            axes[0, 0].legend(
                loc="lower center",
                ncol=4,
                bbox_to_anchor=(0, 0, 1, 1),
                bbox_transform=fig.transFigure,
                fontsize=7.0,
            )

        fig.subplots_adjust(
            top=0.9,
        )
        return fig, axes

    def plot(self, filename, title=None, reciprocal=None, limits=None,
             dtype='rho', return_fig=False, **kwargs):
        """Standard plot of spectrum

        Parameters
        ----------
        filename : str
            Output filename. Include the ending to specify the filetype
            (usually .pdf or .png)
        title : string, optional
            Title for the plot
        reciprocal : :class:`reda.eis.plots.sip_response`, optional
            If another :class:`reda.eis.plots.sip_response` object is provided
            here, use this as the reciprocal spectrum.
        limits : dict, optional
            A dictionary which contains plot limits. See code example below.
        dtype : string, optional
            Determines if the data plotted included geometric factors ('rho')
            or not ('r'). Default: 'rho'
        return_fig : bool, optional
            If True, then do not delete the figure object after saving to file
            and return the figure object. Default: False
        **kwargs : dict
            kwargs is piped through to the _plot function

        Returns
        -------
        fig : :class:`matplotlib.Figure`
            The figure object. Only returned if return_fig is set to True

        Examples
        --------
        >>> from reda.eis.plots import sip_response
        >>> import numpy as np
        >>> frequencies = np.array([
        ...     1.00000000e-03, 1.77827941e-03, 3.16227766e-03, 5.62341325e-03,
        ...     1.00000000e-02, 1.77827941e-02, 3.16227766e-02, 5.62341325e-02,
        ...     1.00000000e-01, 1.77827941e-01, 3.16227766e-01, 5.62341325e-01,
        ...     1.00000000e+00, 1.77827941e+00, 3.16227766e+00, 5.62341325e+00,
        ...     1.00000000e+01, 1.77827941e+01, 3.16227766e+01, 5.62341325e+01,
        ...     1.00000000e+02, 1.77827941e+02, 3.16227766e+02, 5.62341325e+02,
        ...     1.00000000e+03])
        >>> rcomplex = np.array([
        ...     49.34369772-0.51828971j, 49.11781581-0.59248806j,
        ...     48.85819872-0.6331137j , 48.58762806-0.62835135j,
        ...     48.33331113-0.57965851j, 48.11599009-0.50083533j,
        ...     47.94405036-0.41005275j, 47.81528917-0.32210768j,
        ...     47.72215469-0.24543425j, 47.65607773-0.18297794j,
        ...     47.60962191-0.13433101j, 47.57706229-0.09755774j,
        ...     47.55424286-0.07031682j, 47.53822912-0.05041399j,
        ...     47.52697253-0.03601005j, 47.51904718-0.02565412j,
        ...     47.51345965-0.01824266j, 47.50951606-0.01295546j,
        ...     47.50673042-0.00919217j, 47.50476152-0.0065178j ,
        ...     47.50336925-0.00461938j, 47.50238442-0.00327285j,
        ...     47.50168762-0.00231829j, 47.50119454-0.00164187j,
        ...     47.50084556-0.00116268j])
        >>> spectrum = sip_response(frequencies=frequencies, rcomplex=rcomplex)
        >>> fig = spectrum.plot('spectrum.pdf', return_fig=True)

        """
        fig, axes = self._plot(
            reciprocal=reciprocal,
            limits=limits,
            title=title,
            dtype=dtype,
            **kwargs
        )
        fig.savefig(filename, dpi=300)
        if return_fig:
            return fig
        else:
            plt.close(fig)


class multi_sip_response(object):
    """manage multiple sip_response objects and provide some nice overview
    plots
    """
    @staticmethod
    def _is_correct_type(object):
        """check if we can work with this object """
        if not isinstance(object, sip_response):
            raise Exception(
                'can only add sip_reponse.sip_response objects')

    @staticmethod
    def _check_list(object_list):
        if not isinstance(object_list, list):
            raise Exception('can only work with lists')
        [multi_sip_response._is_correct_type(x) for x in object_list]

    def __init__(self, objects=None, labels=None, obj_dict=None):
        """
        Parameters
        ----------
        objects : list|None
            If provided, assume the list to contain multiple spectra in the
            form of sip_response objects
        labels : list|None
            If provided, use the string entries of this list as labels for the
            spectra in objects. Must have the same length as objects, or None.
        obj_dict : dict|None
            Only works if objects is None. Use keys as labels, items as spectra

        """
        # here we store the responses
        if objects is not None:
            multi_sip_response._check_list(objects)
            if len(objects) != len(labels):
                raise Exception(
                    'length of object list must match length of label list')
            self.objects = objects
            self.labels = labels
        elif obj_dict is not None and isinstance(obj_dict, dict):
            self.objects = list(obj_dict.values())
            self._check_list(self.objects)
            self.labels = list(obj_dict.keys())
        else:
            self.objects = []
            self.labels = []
        self.xlim = [None, None]
        self.legend_kwargs = {
            'ncol': 4,
            'fontsize': 6.0,
            'loc': "lower center",
            'bbox_to_anchor': (0, 0, 1, 1),
        }
        self.margin_bottom = 0.3

    def set_xlim(self, xmin, xmax):
        self.xlim = [xmin, xmax]

    def add(self, response, label=None):
        """add one response object to the list
        """
        if not isinstance(response, sip_response):
            raise Exception(
                'can only add sip_reponse.sip_response objects'
            )
        self.objects.append(response)

        if label is None:
            self.labels.append('na')
        else:
            self.labels.append(label)

    def _add_legend(self, ax):
        params = {
            'bbox_transform': ax.get_figure().transFigure,
        }
        params.update(self.legend_kwargs)

        leg = ax.legend(
            **params
        )
        return leg

    def plot_rmag(self, filename, pmin=None, pmax=None, title=None):
        """plot all resistance/resistivity magnitude spectra
        """
        cmap = mpl.cm.get_cmap('viridis')
        SM = mpl.cm.ScalarMappable(norm=None, cmap=cmap)
        colors = SM.to_rgba(np.linspace(0, 1, len(self.objects)))
        fig, ax = plt.subplots(1, 1, figsize=(12 / 2.54, 7 / 2.54))
        for nr, item in enumerate(self.objects):
            ax.semilogx(
                item.frequencies,
                item.rmag,
                '.-',
                color=colors[nr],
                label=self.labels[nr],
            )
        ax.set_ylabel(get_label('rmag', 'meas', 'mathml'))
        ax.set_xlabel('frequency [Hz]')
        ax.set_ylim(pmin, pmax)
        ax.set_xlim(*self.xlim)
        if title is not None:
            ax.set_title(title)
        self._add_legend(ax)
        fig.tight_layout()
        fig.subplots_adjust(bottom=self.margin_bottom)
        fig.savefig(filename, dpi=300, bbox_inches='tight')
        plt.close(fig)

    def plot_rpha(self, filename, pmin=None, pmax=None, title=None):
        """plot all resistance/resistivity phase spectra
        """
        cmap = mpl.cm.get_cmap('viridis')
        SM = mpl.cm.ScalarMappable(norm=None, cmap=cmap)
        colors = SM.to_rgba(np.linspace(0, 1, len(self.objects)))
        fig, ax = plt.subplots(1, 1, figsize=(12 / 2.54, 7 / 2.54))
        for nr, item in enumerate(self.objects):
            ax.semilogx(
                item.frequencies,
                -item.rpha,
                '.-',
                color=colors[nr],
                label=self.labels[nr],
            )
        ax.set_xlim(*self.xlim)
        ax.set_ylabel(get_label('rpha', 'meas', 'mathml'))
        ax.set_xlabel('frequency [Hz]')
        ax.set_ylim(pmin, pmax)
        if title is not None:
            ax.set_title(title)
        self._add_legend(ax)
        fig.tight_layout()
        fig.subplots_adjust(bottom=self.margin_bottom)
        fig.savefig(filename, dpi=300)
        plt.close(fig)

    def plot_cim(self, filename, cmin=None, cmax=None, title=None):
        cmap = mpl.cm.get_cmap('viridis')
        SM = mpl.cm.ScalarMappable(norm=None, cmap=cmap)
        SM = mpl.cm.ScalarMappable(norm=None, cmap=cmap)
        colors = SM.to_rgba(np.linspace(0, 1, len(self.objects)))
        fig, ax = plt.subplots(1, 1, figsize=(12 / 2.54, 7 / 2.54))
        for nr, item in enumerate(self.objects):
            ax.loglog(
                item.frequencies,
                item.cim,
                '.-',
                color=colors[nr],
                label=self.labels[nr],
            )
        ax.set_ylabel(get_label('cim', 'meas', 'mathml'))
        ax.set_xlim(*self.xlim)
        ax.set_xlabel('frequency [Hz]')
        ax.set_ylim(cmin, cmax)
        if title is not None:
            ax.set_title(title)
        self._add_legend(ax)
        fig.tight_layout()
        fig.subplots_adjust(bottom=self.margin_bottom)
        fig.savefig(filename, dpi=300)
        plt.close(fig)

    def plot_cre(self, filename, cmin=None, cmax=None, title=None):
        cmap = mpl.cm.get_cmap('viridis')
        SM = mpl.cm.ScalarMappable(norm=None, cmap=cmap)
        SM = mpl.cm.ScalarMappable(norm=None, cmap=cmap)
        colors = SM.to_rgba(np.linspace(0, 1, len(self.objects)))
        fig, ax = plt.subplots(1, 1, figsize=(12 / 2.54, 7 / 2.54))
        for nr, item in enumerate(self.objects):
            ax.loglog(
                item.frequencies,
                item.cre,
                '.-',
                color=colors[nr],
                label=self.labels[nr],
            )
        ax.set_xlim(*self.xlim)
        ax.set_ylabel(get_label('cre', 'meas', 'mathml'))
        ax.set_xlabel('frequency [Hz]')
        ax.set_ylim(cmin, cmax)
        if title is not None:
            ax.set_title(title)
        self._add_legend(ax)
        fig.tight_layout()
        fig.subplots_adjust(bottom=self.margin_bottom, top=0.9)
        fig.savefig(filename, dpi=300)
        plt.close(fig)



# -*- coding: utf-8 -*-
"""Useful helper functions."""

import os.path
from importlib import import_module

import numpy as np


def opt_import(module, requiredFor="use the full functionality"):
    """Import and return module only if it exists.
    If `module` cannot be imported, a warning is printed followed by the
    `requiredFor` string. Otherwise, the imported `module` will be returned.
    This function should be used to import optional dependencies in order to
    avoid repeated try/except statements.

    Parameters
    ----------
    module : str
        Name of the module to be imported.
    requiredFor : str, optional
        Info string for the purpose of the dependency.

    Examples
    --------
    >>> from reda.utils import opt_import
    >>> reda = opt_import("reda")
    >>> reda.__name__
    'reda'
    >>> opt_import("doesNotExist", requiredFor="do something special")
    No module named 'doesNotExist'.
    You need to install this optional dependency to do something special.
    """
    # set default message for common imports
    if not requiredFor and "crtomo" in module:
        requiredFor = (
            "modelling and inversion with CRTOMO. Check"
            "http://geo.uni-bonn.de/~mweigand/dashboard for details.")

    if not requiredFor and "pygimli" in module:
        requiredFor = ("modelling and inversion with pygimli. Check"
                       "www.pygimli.org for installation details.")

    if module.count(".") > 2:
        raise ImportError("Can only import modules and sub-packages.")

    try:
        mod = import_module(module)
    except ImportError:
        msg = ("No module named \'%s\'.\nYou need to install this optional "
               "dependency to %s.")
        print(msg % (module, requiredFor))
        mod = None

    return mod


def which(program):
    """Python function to mimic the unix 'which' command."""
    # Taken from https://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def search(what):
    """Utility function to search docstrings for string `what`."""
    np.lookfor(what, module="reda", import_modules=False)

def has_multiple_timesteps(data):
    """Return True if `data` container has multiple timesteps."""
    if "timestep" in data.keys():
        if len(np.unique(data["timestep"])) > 1:
            return True
    return False

def split_timesteps(data, consistent_abmn=False):
    """Split data into multiple timesteps."""
    if has_multiple_timesteps(data):
        grouped = data.groupby("timestep")
        return [group[1] for group in grouped]
    else:
        return data



# *-* coding: utf-8 *-*
"""Read MPT DAS-1 data files.

TODO:

"""

import re
import pandas as pd
import numpy as np

from reda.tdip.decay_curve import DecayCurveObj

# from reda.importers.utils.decorators import enable_result_transforms


def get_frequencies(filename, header_row):
    """Read the used frequencies in header of DAS-1 SIP data set.

    Parameters
    ----------
    filename : str
        input filename
    header_row : int
        row number of header row

    Returns
    -------
    frequencies : list
        Contains the measured frequencies

    """

    fid = open(filename, 'r')
    lines = fid.readlines()
    freq_header = lines[header_row + 1].split('Hz')[:-1]
    frequencies = [re.sub(r"[^\d\.]", "", part) for part in freq_header]
    fid.close()
    return frequencies


def import_das1_fd(filename, **kwargs):
    """Reads a frequency domain (single frequency) MPT DAS-1 data file (.Data)
    and pre pares information in pandas DataFrame for further processing.

    Parameters
    ----------
    filename : str
        path to input file
    corr_array : list, optional
        used to correct the electrode numbers [a, b, m, n], eg. for cable
        layouts which separated current and potential cables, hence, 64
        electrodes for a measurement profile of 32 electrodes

    Returns
    -------
    data : :py:class:`pandas.DataFrame`
        Contains the measurement data
    electrodes : :py:class:`pandas.DataFrame`
        Contains electrode positions (None at the moment)
    topography : None
        No topography information is contained in the text files, so we always
        return None


    """
    if 'corr_array' in kwargs:
        corr_array = kwargs.get('corr_array')
    else:
        corr_array = [0, 0, 0, 0]

    df = pd.read_csv(filename,
                     delimiter=' ',
                     comment='!',
                     index_col=0)

    # derive rows used in data block
    data_start = df.index.get_loc('#data_start')
    data_end = df.index.get_loc('#data_end')
    data = df.iloc[data_start + 1: data_end].dropna(axis=1)

    data_new = pd.DataFrame()

    # def split_A(strA):
    #     return int(strA.split(','))

    # A, B, M, N
    data_new['a'] = [
        int(x.split(',')[1]) - corr_array[0] for x in data.iloc[:, 0]]
    data_new['b'] = [
        int(x.split(',')[1]) - corr_array[1] for x in data.iloc[:, 1]]
    data_new['m'] = [
        int(x.split(',')[1]) - corr_array[2] for x in data.iloc[:, 2]]
    data_new['n'] = [
        int(x.split(',')[1]) - corr_array[3] for x in data.iloc[:, 3]]

    data_new['r'] = np.array(data.iloc[:, 4]).astype('float')  # resistance Ohm
    data_new['rpha'] = np.array(data.iloc[:, 5]).astype('float')  # phase mrad
    data_new['I'] = np.array(data.iloc[:, 12]).astype('float')  # current in mA
    data_new['dr'] = np.array(
        data.iloc[:, 9]
    ).astype('float') / (data_new['I'] / 1000)
    data_new['Zt'] = data_new['r'] * np.exp(data_new['rpha'] * 1j / 1000.0)

    datetime_series = pd.to_datetime(data.iloc[:, -7],
                                     format='%Y%m%d_%H%M%S',
                                     errors='ignore')

    data_new['datetime'] = [
        time for index, time in datetime_series.iteritems()
    ]

    return data_new, None, None


def import_das1_td(filename, **kwargs):

    """
    Reads a time domain MPT DAS-1 data file (.Data) and prepares information
    in pandas DataFrame for further processing.

    Parameters
    ----------
    filename : str
        path to input file

    Keyword arguments:
    ------------------
    corr_arry : list
        used to correct the electrode numbers [a, b, m, n], eg. for cable
        layouts which separated current and potential cables, hence, 64
        electrodes for a measurement profile of 32 electrodes

    Returns
    -------
    data : :py:class:`pandas.DataFrame`
        Contains the measurement data
    electrodes : :py:class:`pandas.DataFrame`
        Contains electrode positions (None at the moment)
    topography : None
        No topography information is contained in the text files, so we always
        return None
    """

    if 'corr_array' in kwargs:
        corr_array = kwargs.get('corr_array')
    else:
        corr_array = [0, 0, 0, 0]

    with open(filename, 'r') as fid:
        for idx, line in enumerate(fid):
            if '#data_start' in line:
                d_start = idx
            if '#data_end' in line:
                d_end = idx
            if '!List' in line:
                tm_start = idx
            if '#elec_start' in line:
                tm_end = idx
            if '#TIPDly' in line:
                mdelay = float(line.split('\t')[1])

    # import the data block
    data = pd.read_csv(
        filename,
        delimiter=' ',
        index_col=0,
        names=range(0, 10**3),  # dump the file in huge array
        skiprows=d_start + 1,
        nrows=d_end - d_start - 3,  # skip headers after #data_start
        low_memory=False
    )
    # sometimes there are additional comment lines in the file
    indices_comments = data.index == '!'
    data = data.iloc[~indices_comments]

    # import IPython
    # IPython.embed()

    header = pd.read_csv(
        filename,
        delimiter='\t',
        # comment='!',
        # index_col=0,
        skiprows=tm_start - 1,
        nrows=tm_end - tm_start - 1
    )

    ngates = len(header)
    ipw = np.array(header.iloc[:, 0]).astype(np.float)

    data_new = pd.DataFrame()

    # def split_A(strA):
    #     return int(strA.split(',')[1])

    # data.iloc[:, 0].apply(split_A)

    # A, B, M, N
    data_new['a'] = [
        int(x.split(',')[1]) - corr_array[0] for x in data.iloc[:, 0]]
    data_new['b'] = [
        int(x.split(',')[1]) - corr_array[1] for x in data.iloc[:, 1]]
    data_new['m'] = [
        int(x.split(',')[1]) - corr_array[2] for x in data.iloc[:, 2]]
    data_new['n'] = [
        int(x.split(',')[1]) - corr_array[3] for x in data.iloc[:, 3]]

    data_new['r'] = np.array(data.iloc[:, 4]).astype('float')  # resistance
    data_new['dr'] = np.array(data.iloc[:, 5]).astype('float')  # devR
    # voltage in mV
    data_new['Vab'] = np.array(data.iloc[:, 6].astype('float') * 1000)
    data_new['dVab'] = np.array(
        data.iloc[:, 7].astype('float') * 1000)  # deviation voltage in mV
    # curret in mA
    data_new['I'] = np.array(data.iloc[:, 8 + 2 * ngates]).astype('float')
    data_new['mdelay'] = mdelay

    # use helper DataFrames for Mx, Tm, dMx
    data_m = pd.DataFrame(
        columns=['M' + str(num) for num in range(1, ngates + 1)],
        index=data_new.index
    )
    data_m.loc[:, 'M1':'M' + str(ngates)] = np.array(
        data.iloc[:, 8:8 + 2 * ngates:2]).astype(np.float)  # Mi

    data_tm = pd.DataFrame(
        columns=['Tm' + str(num) for num in range(1, ngates + 1)],
        index=data_new.index
    )
    data_tm.loc[:, 'Tm1':'Tm' + str(ngates)] = ipw

    data_devm = pd.DataFrame(
        columns=['devm' + str(num) for num in range(1, ngates + 1)],
        index=data_new.index
    )
    data_devm.loc[:, 'devm1':'devm' + str(ngates)] = np.array(
        data.iloc[:, 9:9 + 2 * ngates:2]
    ).astype(np.float)  # devMi

    # compute the global chargeability
    nominator = np.sum(
        np.array(data_m.loc[:, 'M1': 'M' + str(ngates)]) *
        np.array(data_tm.loc[:, 'Tm1': 'Tm' + str(ngates)]),
        axis=1
    )
    denominator = np.sum(
        np.array(data_tm.loc[:, 'Tm1': 'Tm' + str(ngates)]),
        axis=1
    )
    data_new['chargeability'] = nominator / denominator

    datetime_series = pd.to_datetime(
        data.iloc[:, 10 + 2 * ngates],
        format='%Y%m%d_%H%M%S',
        errors='ignore'
    )

    data_new['datetime'] = [
        time for index, time in datetime_series.iteritems()]

    data_new['decayCurve'] = 0
    # construct a sub DataFrame for decay curve properties
    for index, meas in data_m.iterrows():
        decaycurve = pd.DataFrame(
            index=range(len(ipw)), columns=['Mx', 'T[ms]', 'dMx']
        )
        decaycurve['Mx'] = meas.values
        # use the gate ending as plotting point
        decaycurve['T[ms]'] = mdelay + np.cumsum(ipw)
        decaycurve['dMx'] = data_devm.iloc[index, :].values
        decaycurve = decaycurve.set_index('T[ms]')
        data_new.at[index, 'decayCurve'] = DecayCurveObj(decaycurve)

    return data_new, None, None


def import_das1_sip(filename, **kwargs):
    """Reads a spectral induced polarization MPT DAS-1 data file (.Data) and
    prepares information in pandas DataFrame for further processing.

    Parameters
    ----------
    filename : string
        path to input file

    Keyword arguments:
    ------------------
    corr_arry : list
        used to correct the electrode numbers [a, b, m, n], eg. for cable
        layouts which separated current and potential cables, hence, 64
        electrodes for a measurement profile of 32 electrodes

    Returns
    -------
    data : :py:class:`pandas.DataFrame`
        Contains the measurement data
    electrodes : :py:class:`pandas.DataFrame`
        Contains electrode positions (None at the moment)
    topography : None
        No topography information is contained in the text files, so we always
        return None
    """

    if 'corr_array' in kwargs:
        corr_array = kwargs.get('corr_array')
    else:
        corr_array = [0, 0, 0, 0]

    d_start, d_end = 0, 0
    # deduce the data block here
    # look for lines with #data_start/#data_end

    with open(filename, 'r') as fid:
        for idx, line in enumerate(fid):
            if '#data_start' in line:
                d_start = idx
            if '#data_end' in line:
                d_end = idx

    # import the data block
    data = pd.read_csv(
        filename,
        delimiter=' ',
        index_col=0,
        names=range(0, 10 ** 3),  # dump the file in huge array
        skiprows=d_start + 3,
        # skip headers after #data_start
        nrows=d_end - d_start - 4,
        low_memory=False
    )
    # sometimes there are additional comment lines in the file
    indices_comments = data.index == '!'
    data = data.iloc[~indices_comments]

    frequency_list = get_frequencies(filename, d_start)
    frequencies = np.array(frequency_list).astype(float)
    num_freqs = len(frequencies)
    num_meas = d_end - d_start - 4

    # number of nan and unused columns present when quadrupoles has
    # << * * TX Resist. out of range * * >> error
    tx_out_skip = 22
    nan_index = None
    # identifier if quadrupole has above mentioned error
    fskip_count = np.zeros(num_meas).astype(int)

    data_new = pd.DataFrame()
    data_new['a'] = [
        int(x.split(',')[1]) - corr_array[0] for x in data.iloc[:, 0]]
    data_new['b'] = [
        int(x.split(',')[1]) - corr_array[1] for x in data.iloc[:, 1]]
    data_new['m'] = [
        int(x.split(',')[1]) - corr_array[2] for x in data.iloc[:, 2]]
    data_new['n'] = [
        int(x.split(',')[1]) - corr_array[3] for x in data.iloc[:, 3]]
    data_fin = pd.DataFrame(
        columns=[
            'a', 'b', 'm', 'n', 'frequency', 'Zt', 'r', 'dr', 'rpha', 'drpha',
            'I', 'datetime'
        ]
    )

    # array to check for error
    dout_r = np.zeros((num_meas, len(frequencies)))

    for idx, freq in enumerate(frequencies):
        print('Processing frequency: %s Hz' % str(freq))

        if nan_index is not None:

            # iterate over quadrupoles and skip columns based on fskip_count
            for row_idx in range(len(data)):
                data_new.loc[row_idx, 'frequency'] = freq
                # resistance
                data_new.loc[row_idx, 'r'] = float(
                    data.iloc[
                        row_idx,
                        idx * 6 + 4 + tx_out_skip * fskip_count[row_idx]
                    ]
                )
                # phase
                data_new.loc[row_idx, 'rpha'] = float(
                    data.iloc[
                        row_idx, idx * 6 + 6 + tx_out_skip * fskip_count[
                            row_idx]])
                # current in mA
                data_new.loc[row_idx, 'I'] = float(
                    data.iloc[
                        row_idx, idx * 6 + 8 +
                        tx_out_skip * fskip_count[row_idx]])
                # devR
                data_new.loc[row_idx, 'dr'] = float(
                    data.iloc[
                        row_idx, idx * 6 + 5 + tx_out_skip * fskip_count[
                            row_idx]])
                # devPhi
                data_new.loc[row_idx, 'drpha'] = float(
                    data.iloc[
                        row_idx, idx * 6 + 7 + tx_out_skip * fskip_count[
                            row_idx]])
                # array to check for error
                dout_r[row_idx, idx] = float(
                    data.iloc[
                        row_idx, idx * 6 + 4 + tx_out_skip *
                        fskip_count[row_idx]])

        else:

            dout_r[:, idx] = np.array(data.iloc[:, idx * 6 + 4])
            data_new['frequency'] = freq
            # resistance
            data_new['r'] = np.array(data.iloc[:, idx * 6 + 4]).astype('float')
            # phase
            data_new['rpha'] = np.array(
                data.iloc[:, idx * 6 + 6]).astype('float')
            # current in mA
            data_new['I'] = np.array(data.iloc[:, idx * 6 + 8]).astype('float')
            # devR
            data_new['dr'] = np.array(data.iloc[:, idx * 6 + 5]).astype(
                'float')
            # dev Phi
            data_new['drpha'] = np.array(
                data.iloc[:, idx * 6 + 7]).astype('float')

        # check for quadrupoles containing nans (because of error)
        nan_index = np.where(np.isnan(dout_r[:, idx]) == 1)[0]
        # set the skip count
        fskip_count[nan_index] = fskip_count[nan_index] + 1
        fskip_count.astype(int)

        data_fin = data_fin.append(data_new, ignore_index=True, sort=False)

    # compute Zt
    data_fin['Zt'] = data_fin['r'] * np.exp(data_fin['rpha'] * 1j / 1000.0)

    #
    start = len(frequencies) - 1
    for row_idx in range(len(data)):
        data_new.at[row_idx, 'datetime'] = data.iloc[
            row_idx, start * 6 + 11 + tx_out_skip * fskip_count[row_idx]]

    datetime_series = pd.to_datetime(
        data_new['datetime'], format='%Y%m%d_%H%M%S', errors='ignore')
    datetime_stack = datetime_series.append(
        [datetime_series] * (num_freqs - 1)
    )
    datetime_stack_reindex = datetime_stack.reset_index()
    data_fin['datetime'] = datetime_stack_reindex['datetime']

    data_fin[['a', 'b', 'm', 'n']] = data_fin[['a', 'b', 'm', 'n']].astype(
        int
    )

    return data_fin, None, None


def get_measurement_type(filename):
    """
    Given an MPT DAS-1 result file, try to determine the type of measurement.

    Currently supported/detected types:

    * TDIP (i.e., time-domain measurement)
    * FD (i.e., complex measurement)
    * SIP (i.e., sEIT measurements

    Parameters
    ----------
    filename : str
        Path to data file

    Returns
    -------
    measurement_type : str
        The type of measurement: tdip,cr,sip

    """
    with open(filename, 'r') as fid:
        for line in fid:
            if '!TDIP' in line:
                return 'tdip'
            if '!Spectral' in line:
                return 'sip'
            if '!FDIP' in line:
                return 'cr'

    raise Exception(
        'Data type (FD/TD/SIP) of '
        '{0} cannot be determined. Check the file!'.format(filename)
    )


def import_das1(filename, **kwargs):

    """
    Reads a any MPT DAS-1 data file (.Data), e.g. TD/FD/SIP, and prepares
    information in pandas DataFrame for further processing.

    Parameters
    ----------
    filename : string
        path to input file

    Keyword arguments:
    ------------------
    corr_arry : list
        used to correct the electrode numbers [a, b, m, n], eg. for cable
        layouts which separated current and potential cables, hence, 64
        electrodes for a measurement profile of 32 electrodes

    Returns
    -------
    data : :py:class:`pandas.DataFrame`
        Contains the measurement data
    electrodes : :py:class:`pandas.DataFrame`
        Contains electrode positions (None at the moment)
    topography : None
        No topography information is contained in the text files, so we always
        return None
    """
    measurement_type = get_measurement_type(filename)
    if measurement_type == 'tdip':
        data, electrodes, topography = import_das1_td(
            filename, **kwargs)
        return data, electrodes, topography
    elif measurement_type == 'sip':
        data, electrodes, topography = import_das1_sip(
            filename, **kwargs)
        return data, electrodes, topography
    elif measurement_type == 'cr':
        data, electrodes, topography = import_das1_fd(
            filename, **kwargs)
        return data, electrodes, topography



import os
import logging

# import h5py
# import numpy as np
import pandas as pd

from reda.exporters.tsert_export import tsert_base


class tsert_import(tsert_base):
    """The TSERT file format -- export functions

    TSERT: Time-series electrical resistivity tomography
    """
    def __init__(self, filename):
        assert os.path.isfile(filename), "File must exist"
        super().__init__(filename)

        # import available timesteps
        # f = h5py.File(filename, 'r')
        # self.available_timesteps = list(f['ERT_DATA'].keys())
        # f.close()

    def _setup_logger(self):
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )

        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger

        logger = logging.getLogger(__name__)
        print('TODO: Fix logging format')
        # while logger.hasHandlers():
        #     logger.removeHandler(logger.handlers[0])
        # logger.addHandler(ch)
        logger.setLevel(logging.INFO)

        self.logger = logger

    # def _open_file(self, mode='r'):
    #     if self.fid is None or not self.fid or mode != self.fid_mode:
    #         if self.fid:
    #             self.fid.close()
    #         self.fid = h5py.File(self.filename, mode)
    #         self.fid_mode = mode
    #     return self.fid

    # def _close_file(self):
    #     if self.fid:
    #         self.fid.close()
    #         self.fid_mode = None
    #         self.fid = None

    def _check_hdf_file(self):
        """Apply various checks to the registered file

        """
        f = self._open_file('r')
        assert 'ERT_DATA' in f.keys()

        print('TODO: Check all timesteps for presence of base-version')

    # def load_all_timesteps(self, version):
    #     data_list = []
    #     for timestep in self.available_timesteps:
    #         key = 'ERT_DATA/{}/{}'.format(timestep, version)
    #         print(key)
    #         # TODO check if key exists
    #         data = pd.read_hdf(self.filename, key)
    #         data_list.append(data)
    #     return data_list

    def import_data(
            self, timesteps, version='base', not_before=None, not_after=None):
        """Wrapper for .load_data that merges the data into one dataframe
        """
        data_list = self.load_data(
            timesteps,
            version,
            not_before,
            not_after,
        )
        data = pd.concat(data_list)
        return data

    def load_data(
            self, timesteps='all', version='base',
            not_before=None, not_after=None):
        """
        Parameters
        ----------
        timesteps : iterable|'all'
            Timesteps to import, the string 'all' will import all data
        version :

        not_before : None|datetime.datetime
            Only select data sets whose timestep lies before the given
            datetime.  Only evaluated if the timesteps parameter is set to
            'all'.
        not_after : None|datetime.datetime
            Only select data sets whose timestep lies after the given datetime.
            Only evaluated if the timesteps parameter is set to 'all'.

        Returns
        -------
        data_list : list
            List of dataframes

        """
        data_index = self.get_index()
        if data_index.shape[0] == 0:
            # no data present
            print('No data in file')
            return

        # import IPython
        # IPython.embed()
        index_is_datetime = False
        if pd.api.types.is_datetime64_any_dtype(data_index["value"].dtype):
            index_is_datetime = True
        print('DTYPES', data_index.dtypes)

        # all keys
        ts_keys = data_index.index.values
        reversed_data_index = data_index.reset_index().set_index(
            'value').sort_index()

        if timesteps == 'all':
            if not_before is not None or not_after is not None:
                print('TRUNCATING')
                if not index_is_datetime:
                    raise Exception('before/after requires datetime indices')
                ts_keys = reversed_data_index.truncate(
                    before=not_before,
                    after=not_after,
                )['index']
                print('TS_KEYS filtered', ts_keys)
        else:
            assert timesteps in data_index['value'].values
            ts_keys = [
                data_index.reset_index().set_index('value').loc[timesteps],
            ]

        print('loading version:', version)

        data_list = {}
        for ts_key in ts_keys:
            key = 'ERT_DATA/{}/{}'.format(ts_key, version)
            print(key)
            # TODO check if key exists
            data = pd.read_hdf(self.filename, key)
            timestep = data_index.loc[ts_key, 'value']
            print('timestep', timestep)
            data_list[timestep] = data
        return data_list

    def load_topography(self):
        self._open_file()
        key = '/TOPOGRAPHY/topography'
        if key in self.fid:
            topo_is_present = True
            self.logger.info('Loading topography from file')
        else:
            self.logger.info('No topography present in file')
            topo_is_present = False
        self._close_file()

        if topo_is_present:
            topography = pd.read_hdf(self.filename, key)
            return topography
        return None

    def load_electrode_positions(self):
        self._open_file()
        key = '/ELECTRODES/ELECTRODE_POSITIONS'
        if key in self.fid:
            elec_coords_are_present = True
            self.logger.info('Loading electrode_positions from file')
        else:
            self.logger.info('No electrode_positions present in file')
            elec_coords_are_present = False
        self._close_file()

        if elec_coords_are_present:
            electrode_positions = pd.read_hdf(self.filename, key)
            return electrode_positions
        return None

    def summary(self, print_index=False):
        """Short summary of the filename

        Parameters
        ----------
        print_index : bool, optional
        """
        print(80 * '#')
        print('Summary of file: {}'.format(self.filename))
        f = self._open_file('r')
        print('Format metadata:')
        print('File format:', f.attrs['file_format'])
        print('Format version:', f.attrs['format_version'])
        print('-' * 60)
        nr_timesteps = len(f['ERT_DATA'].keys())
        print('Number of time steps: {}'.format(nr_timesteps))
        # print the first ten timesteps:
        timesteps = list(f['ERT_DATA'].keys())
        for i in range(0, min(10, nr_timesteps)):
            print('{}: /ERT_DATA/{}'.format(i + 1, timesteps[i]))

        # go through all timesteps and enumerate the versions
        versions = {}
        for timestep in f['ERT_DATA'].keys():
            for version in f['ERT_DATA'][timestep]:
                versions[version] = versions.get(version, 0) + 1
        print('Available versions:')
        for version, numbers in versions.items():
            print(
                ' - Version {} is present {} times'.format(version, numbers)
            )
        print(80 * '#')
        f.close()

        if print_index:
            print('-----------------')
            print('Index dataframe:')
            index = pd.read_hdf(
                self.filename,
                '/INDEX/index',
            )
            print(index)
            print('-----------------')

    def _load_metadata_from_group(self, f, base):
        metadata = {}
        for key, item in f[base].attrs.items():
            metadata[key] = item

        for subgroup in f[base].keys():
            new_base = base + '/' + subgroup
            metadata[subgroup] = self._load_metadata_from_group(
                f, new_base
            )

        return metadata

    def load_metadata(self):
        f = self._open_file('r')
        assert 'METADATA' in f.keys(), "key METADATA not present!"
        metadata = self._load_metadata_from_group(f, 'METADATA')

        self._close_file()
        return metadata
